# npchase.py example game
# by NpChase and Jacob Wolf

from quest.contrib.npchase import Chase
from quest.examples.grandmas_soup import GrandmasSoupGame


class AffectionateGrandmaGame(GrandmasSoupGame):
    """Same game as GrandmasSoup but in this version she REALLY wants to talk to you.
    """

    def setup_npcs(self):
        """ Sets up the NPCs and gives grandma a chase strategy
        """
        super().setup_npcs()
        grandma = self.npc_list[0]
        walk = Chase()
        grandma.strategy = walk

if __name__ == '__main__':
    game = AffectionateGrandmaGame() 
    game.run()




# sprite_directionality.py example game
# by Team Sonic and Jacob Wolf
#
# This example game showcases the sprite directionality feature found in
# quest/contrib/sprite_directionality.py
#
# Built based on code by Paul Vicent Craven

import arcade

TEXTURE_LEFT = 0
TEXTURE_RIGHT = 1
TEXTURE_DOWN = 2
TEXTURE_UP = 3

class DirectionalMixin:
    """ A mixin for `quest.sprite.Sprite` that allows the sprite to face different
    directions depending on which direction the sprite is moving.

    Arguments:
        player_sprite_image_lr (string): file path to image for left and
                                         right player sprite image
        player_sprite_image_up (string): file path to up facing image
        player_sprite_image_down (string): file path to down facing image
        player_scaling (int): scale for the sprite image
    """
    def __init__(self, player_sprite_image_lr, player_sprite_image_up, player_sprite_image_down, player_scaling):
        super().__init__(player_sprite_image_lr, player_scaling)
        self.textures = []
        texture_left = arcade.load_texture(player_sprite_image_lr, mirrored=True)
        self.textures.append(texture_left)
        texture_right = arcade.load_texture(player_sprite_image_lr)
        self.textures.append(texture_right)
        texture_down = arcade.load_texture(player_sprite_image_down)
        self.textures.append(texture_down)
        texture_up = arcade.load_texture(player_sprite_image_up)
        self.textures.append(texture_up)
        
        self.set_texture(TEXTURE_RIGHT)

    def on_update(self, game):
        """Change texture based on player direction.
        """
        super().on_update(game)
        if self.change_x < 0:
            self.texture = self.textures[TEXTURE_LEFT]
        elif self.change_x > 0:
            self.texture = self.textures[TEXTURE_RIGHT]
        if self.change_y < 0:
            self.texture = self.textures[TEXTURE_DOWN]
        elif self.change_y > 0:
            self.texture = self.textures[TEXTURE_UP]




import arcade
from quest.helpers import tint, shade
from textwrap import wrap

class TextLabel:
    """Draws text onto a specific part of the screen. 
    """
    width = 400
    font_size = 16
    line_height = 20
    padding = 6
    font_color = arcade.color.BLACK
    background_color = arcade.color.LIGHT_GRAY
    wrap_at = 40
    highlight = False

    def __init__(self, text, x_center, y_center):
        self.x_center = x_center
        self.y_center = y_center
        if self.wrap_at:
            self.text_lines = wrap(text, self.wrap_at)
        else:
            self.text_lines = [text]

    def draw(self):
        if self.background_color:
            arcade.draw_rectangle_filled(
                self.x_center, 
                self.y_center, 
                self.width,
                self.height(),
                tint(self.background_color) if self.highlight else self.background_color 
            )
        for i, line in enumerate(self.text_lines):
            arcade.draw_text(
                line, 
                self.text_x(),
                self.text_y(i),
                tint(self.font_color) if self.highlight else self.font_color,
                self.font_size
            )

    def height(self):
        return len(self.text_lines) * self.line_height + 2 * self.padding

    def text_x(self):
        return self.x_center - self.width / 2 + self.padding
    
    def text_y(self, line=0):
        y_top = self.y_center + self.height() / 2 
        return y_top - (line + 1) * (self.padding + self.line_height)

class TextLabelStack:
    """ Creates a stack of TextLabels
    """
    text_label_class = TextLabel
    margin = 10

    def __init__(self, text_for_labels, x_center, y_top):
        self.x_center = x_center
        self.y_top = y_top
        self.text_labels = []
        for i, text in enumerate(text_for_labels):
            label = self.text_label_class(
                text, 
                self.x_center,
                self.y_top - self.height()
            )
            label.y_center -= label.height() / 2
            self.text_labels.append(label)

    def height(self):
        label_heights = sum([label.height() for label in self.text_labels])
        margin_heights = (1 + len(self.text_labels)) * self.margin
        return label_heights + margin_heights

    def draw(self):
        for label in self.text_labels:
            label.draw()

    def get_highlight(self):
        for label_index, label in enumerate(self.text_labels):
            if label.highlight:
                return label_index

    def set_highlight(self, i):
        for label_index, label in enumerate(self.text_labels):
            label.highlight = (i == label_index)

    def __len__(self):
        return len(self.text_labels)



from quest.game import QuestGame
from quest.map import TiledMap
from quest.dialogue import Dialogue
from quest.modal import Modal, DialogueModal
from quest.sprite import QuestSprite, Player, Wall, NPC
from quest.helpers import scale, resolve_resource_path
from quest.strategy import RandomWalk
import arcade
import os
from pathlib import Path

class GrandmasSoupGame(QuestGame):
    """Help Grandma find all the ingredients for her soup.

    :py:class:`GrandmasSoupGame` shows off the full range of features in the Quest
    framework, loading a map and letting the player explore it. Grandma and the Vegetables
    are subclasses of :py:class:`NPC`. A :py:class:`quest.modal.DialogueModal` is used for the conversations
    with Grandma. To run this example, run::

        $ python -m quest.examples.grandmas_soup

    After you play it, check out the sorce code by clicking on "source" in the
    blue bar just above.

    Attributes:
        dialogue: A :py:class:`Dialogue` containing the game's conversation. 
        modal: A :py:class:`DialogueModal` to show the dialogue.
        items: A list to keep track of which items the player has collected.
    """

    player_sprite_image = resolve_resource_path("images/boy_simple.png")
    screen_width = 500
    screen_height = 500
    left_viewport_margin = 96                            
    right_viewport_margin = 96
    bottom_viewport_margin = 96
    top_viewport_margin = 96
    player_initial_x = 300
    player_initial_y = 300
    player_speed = 8

    def __init__(self):
        super().__init__()
        self.dialogue = Dialogue.from_ink(resolve_resource_path("grandma.ink"))
        self.modal = DialogueModal(self, self.dialogue)
        self.items = []

    def setup_maps(self):
        """Sets up the standard island map.
        """
        super().setup_maps()
        sprite_classes = {
            "Obstacles": Wall,
            "Background": QuestSprite,
        }
        self.add_map(TiledMap(resolve_resource_path("images/island/island.tmx"), sprite_classes))

    def setup_walls(self):
        """As in other examples, assigns all sprites in the "Obstacles" layer to be walls.
        """
        self.wall_list = self.get_current_map().get_layer_by_name("Obstacles").sprite_list

    def setup_npcs(self):
        """Creates and places Grandma and the vegetables.
        """
        npc_data = [
            [Grandma, "images/people/grandma.png", 3, 400, 400],
            [Carrots, "images/items/carrots.png", 1, 220, 640],
            [Mushroom, "images/items/mushroom.png", 1, 1028, 264],
            [Potatoes, "images/items/potatoes.png", 1, 959, 991],
            [Tomatos, "images/items/tomatos.png", 1, 323, 1055],
        ]
        self.npc_list = arcade.SpriteList()
        for sprite_class, image, scale, x, y in npc_data:
            sprite = sprite_class(resolve_resource_path(image), scale)
            sprite.center_x = x
            sprite.center_y = y
            self.npc_list.append(sprite)

        grandma = self.npc_list[0]
        walk = RandomWalk(0.05)
        grandma.strategy = walk

    def talk_with_grandma(self):
        """Opens the dialogue modal to show conversation with Grandma. This is called
        when the player collides with Grandma.
        """
        self.open_modal(self.modal)

    def got_item(self, description):
        """Adds the item's description to the items list. This is called when the 
        player collides with a vegetable.

        Arguments:
            description: The item description.
        """
        self.items.append(description.upper())
        if len(self.items) < 4:
            self.dialogue.run(self.items[-1])
        else:
            self.dialogue.run("SOUP")

class Grandma(NPC):
    """Grandma is an NPC. 

    Attributes:
        repel_distance: How far back the player should be pushed after colliding
            with Grandma. This is necessary because otherwise when the dialogue modal 
            closed, it would immediately reopen. Grandma is interesting, but not that 
            interesting.
    """
    repel_distance = 20

    def on_collision(self, sprite, game):
        """When the player collides with Grandma, she repels the player and then 
        :py:meth:`talk_with_grandma` is called to open the dialogue modal.
        """
        self.repel(sprite)
        if isinstance(sprite, Player):
            game.talk_with_grandma()

    def repel(self, sprite):
        "Backs the sprite away from self"
        away = (self.center_x - sprite.center_x, self.center_y - sprite.center_y)
        away_x, away_y = scale(away, self.repel_distance)
        sprite.center_x = sprite.center_x - away_x
        sprite.center_y = sprite.center_y - away_y
        sprite.stop()

class Vegetable(NPC):
    """A vegetable is an NPC that can be picked up.
    """
    description = "item"
    def on_collision(self, sprite, game):
        """When the player collides with a vegetable, it tells the game and then 
        kills itself.
        """
        game.got_item(self.description)
        self.kill()

class Carrots(Vegetable):
    description = "carrots"

class Mushroom(Vegetable):
    description = "mushroom"

class Potatoes(Vegetable):
    description = "potatoes"

class Tomatos(Vegetable):
    description = "tomatos"

if __name__ == '__main__':
    game = GrandmasSoupGame()
    game.run()



import arcade
from arcade.sprite_list import SpriteList, _check_for_collision
from arcade import check_for_collision_with_list
from itertools import chain, combinations
from collections import defaultdict
from easing_functions import LinearInOut
from quest.helpers import Direction, SpriteListList, scale
from time import time
from math import sqrt

class QuestPhysicsEngine:
    """Base class for Quest Physics Engines

    The engine is initialized with a :py:class:`QuestGame` instance, which
    the engine uses to access sprites. Quest's physics engines make some assumptions
    about the structure of a game in order to simplify logic. It is assumed that the
    game has attributes `player_list`, `wall_list`, and `npc_list`. Players and NPCs will
    be moved according to their `change_x` and `change_y` attributes; walls will not move.
    When players or NPCs collide with walls, they are pushed back until they are no longer
    colliding. When players or NPCs collide with each other, their `on_collision` methods
    are called, but they are not automatically repelled. If you want players or NPCs to be
    repelled from each other, see :py:class:`quest.examples.grandmas_soup.Grandma`. 

    Args:
        game (QuestGame): The game to which the engine will be attached.
    """

    def __init__(self, game):
        self.game = game
        self.player_list=game.player_list
        self.wall_list=game.wall_list
        self.npc_list=game.npc_list
        all_sprite_lists = [self.player_list, self.wall_list, self.npc_list]
        curr_map = self.game.get_current_map()
        for layer in curr_map.layers:
            all_sprite_lists.append(layer.sprite_list)
        self.all_sprites = SpriteListList(all_sprite_lists)

    def update(self, game):
        """waits for implementation from a more complex physics engine like py:class:`ContinuousPhysicsEngine`
        """

    def player(self):
        return self.player_list.sprite_list[0]

class ContinuousPhysicsEngine(QuestPhysicsEngine):
    """A continuous physics engine allows sprites to be at any point.

    This implementation is simple but inefficient. It may be problematic with
    more complex games. If we run into trouble, first try using spatial hashes
    to resolve collisions.
    """
    def __init__(self, game, **kwargs):
        super().__init__(game, **kwargs)
        self.non_wall_list = SpriteListList([self.player_list, self.npc_list])

    def update(self):
        """Updates sprite positions and handles collisions.
        """
        super().update(self.game)
        self.update_sprite_positions()
        self.resolve_collisions_with_walls()
        self.resolve_collisions_between_nonwalls()

    def update_sprite_positions(self):
        """Updates sprite positions using their `change_x` and `change_y` attributes.
        """
        for moving_sprite in self.non_wall_list:
            moving_sprite.center_x += moving_sprite.change_x
            moving_sprite.center_y += moving_sprite.change_y

    def resolve_collisions_with_walls(self):
        """Resolves collisions between every sprite and every wall.
        """
        for moving_sprite in self.non_wall_list:
            if moving_sprite.change_x == 0 and moving_sprite.change_y == 0:
                continue
            wall_collisions = check_for_collision_with_list(moving_sprite, self.wall_list)
            for wall in wall_collisions:
                moving_sprite.on_collision(wall, self.game)
                wall.on_collision(moving_sprite, self.game)
                self.resolve_sprite_wall_collision(moving_sprite, wall)

    def resolve_sprite_wall_collision(self, sprite, wall):
        """Stops the sprite and backs it away from the wall until they are no longer colliding.

        The distance by which the sprite backs up doubles until they are no longer colliding.
        Note that this does not handle the case in which backing away from one wall means
        it hits another wall (e.g. in a narrow passageway). This will be handled on the subsequent
        update. We can write more complex code if it becomes necessary.
        """
        sprite.stop()
        repel_distance = 1
        away = (wall.center_x - sprite.center_x, wall.center_y - sprite.center_y)
        while _check_for_collision(sprite, wall):
            away_x, away_y = scale(away, repel_distance)
            sprite.center_x = sprite.center_x - away_x
            sprite.center_y = sprite.center_y - away_y
            repel_distance *= 2

    def resolve_collisions_between_nonwalls(self):
        """For every pair of nonwall sprites, resolves collisions.
        """
        for sprite0, sprite1 in combinations(self.non_wall_list, 2):
            if _check_for_collision(sprite0, sprite1):
                sprite0.on_collision(sprite1, self.game)
                sprite1.on_collision(sprite0, self.game)

class DiscretePhysicsEngine(QuestPhysicsEngine):
    """A physics engine which snaps sprite movement to specific gridpoints.

    Some games work better when sprites occupy specific tiles, rather than
    having continuous positions. This can make it easier to think about
    relationships between sprites (for example, to calculate which are
    adjacent, or to plan a route). :py:class:`DiscretePhysicsEngine` handles
    sprite movement in a discrete way, while animating sprites' transitions
    from tile to tile.

    Args:
        player_sprite (arcade.Sprite):
        dynamic_sprite_lists (bool): Whether new sprites might be added
            to sprite lists during the game. Performance is better when
            False. Default True.

    Attributes:
        tile_transition_cutoff (float): Parametric t value at which a sprite's
            current tile should shift to the

    """

    tile_transition_cutoff = 0.5
    easing_class = LinearInOut

    def __init__(self, game, grid_map_layer, diagonal=True, check_for_new_sprites=True, **kwargs):
        super().__init__(game, **kwargs)
        self.grid = grid_map_layer
        self.diagonal = diagonal
        self.easing = self.easing_class()
        self.check_for_new_sprites = check_for_new_sprites
        sprite_lists = [self.player_list, self.wall_list, self.npc_list]
        self.all_nonbackground_sprites = SpriteListList(sprite_lists)
        self.dynamic_sprites = SpriteListList([l for l in sprite_lists if not l.is_static])
        self.ensure_sprite_metadata(all_sprites=True)
        self.tile_positions = defaultdict(list)
        for sprite in self.all_nonbackground_sprites:
            self.tile_positions[sprite.current_tile].append(sprite)

    def update(self):
        super().update(self.game)
        if self.check_for_new_sprites:
            self.ensure_sprite_metadata()
        for sprite in self.dynamic_sprites:
            if (sprite.change_x or sprite.change_y) and not sprite.moving:
                self.begin_move(sprite)
            elif sprite.moving:
                self.continue_move(sprite)

    def begin_move(self, sprite):
        direction = Direction.from_vector((sprite.change_x, sprite.change_y), self.diagonal)
        ox, oy = sprite.origin_tile
        vx, vy = direction.to_vector()
        destination = (ox + vx, oy + vy)
        if self.grid.position_in_grid(destination):
            wall = self.get_wall(self.tile_positions[destination])
            if wall:
                self.player().on_collision(wall, self.game)
                wall.on_collision(self.player(), self.game)
            else:
                sprite.moving = True
                sprite.move_start = time()
                sprite.t = 0
                sprite.move_direction = direction
                sprite.destination_tile = destination

    def continue_move(self, sprite):
        duration = 1 / sprite.speed
        if sprite.move_direction.is_diagonal():
            duration *= sqrt(2)
        sprite.t = self.ease((time() - sprite.move_start) / duration)
        if sprite.t >= self.tile_transition_cutoff:
            self.tile_positions[sprite.current_tile].remove(sprite)
            sprite.current_tile = sprite.destination_tile
            for other_sprite in self.tile_positions[sprite.current_tile]:
                sprite.on_collision(other_sprite, self.game)
                other_sprite.on_collision(sprite, self.game)
            self.tile_positions[sprite.current_tile].append(sprite)
        if sprite.t >= 1.0:
            self.end_move(sprite)
        else:
            p0 = self.grid.get_pixel_position(sprite.origin_tile)
            p1 = self.grid.get_pixel_position(sprite.destination_tile)
            sprite.center_x, sprite.center_y = self.interpolate(p0, p1, sprite.t)

    def end_move(self, sprite):
        sprite.moving = False
        sprite.center_x, sprite.center_y = self.grid.get_pixel_position(sprite.destination_tile)
        sprite.origin_tile = sprite.current_tile = sprite.destination_tile

    def get_wall(self, sprite_list):
        for sprite in sprite_list:
            if sprite in self.wall_list:
                return sprite

    def interpolate(self, p0, p1, t):
        (x0, y0), (x1, y1) = p0, p1
        return x0 + (x1 - x0) * t, y0 + (y1 - y0) * t

    def ease(self, x):
        return self.easing.ease(x)

    def ensure_sprite_metadata(self, all_sprites=False):
        sprites = self.all_nonbackground_sprites if all_sprites else self.dynamic_sprites
        for sprite in sprites:
            if not hasattr(sprite, 'current_tile'):
                sprite.origin_tile = self.grid.get_grid_position((sprite.center_x, sprite.center_y))
                sprite.current_tile = sprite.origin_tile
                sprite.destination_tile = None
                sprite.move_start = None
                sprite.moving = False
                sprite.t = None



# Really horrible stuff in here. Don't look.
# Based on arcade v2.3.8

import os
from arcade import SpriteList
import pytiled_parser
from arcade.tilemap import (
    get_tilemap_layer,
    _get_tile_by_gid,
    _get_image_source,
    _get_image_info_from_tileset,
)

def process_layer(sprite_class, map_object, layer_name, scaling=1, base_directory=""):
    if len(base_directory) > 0 and not base_directory.endswith("/"):
        base_directory += "/"
    layer = get_tilemap_layer(map_object, layer_name)
    if layer is None:
        print(f"Warning, no layer named '{layer_name}'.")
        return SpriteList()
    if isinstance(layer, pytiled_parser.objects.TileLayer):
        return _process_tile_layer(sprite_class, map_object, layer, scaling, base_directory)
    elif isinstance(layer, pytiled_parser.objects.ObjectLayer):
        raise NotImplementedError("This shim does not include _process_object_layer")
    print(f"Warning, layer '{layer_name}' has unexpected type. '{type(layer)}'")
    return SpriteList()

def _process_tile_layer(sprite_class, map_object, layer, scaling=1, base_directory=""):
    sprite_list = SpriteList()
    map_array = layer.data
    for row_index, row in enumerate(map_array):
        for column_index, item in enumerate(row):
            if item == 0:
                continue
            tile = _get_tile_by_gid(map_object, item)
            if tile is None:
                print(f"Warning, couldn't find tile for item {item} in layer "
                      f"'{layer.name}' in file '{map_object.tmx_file}'.")
                continue
            my_sprite = _create_sprite_from_tile(sprite_class, map_object, tile, scaling=scaling,
                                                 base_directory=base_directory)
            if my_sprite is None:
                print(f"Warning: Could not create sprite number {item} in layer '{layer.name}' {tile.image.source}")
            else:
                my_sprite.center_x = column_index * (map_object.tile_size[0] * scaling) + my_sprite.width / 2
                my_sprite.center_y = (map_object.map_size.height - row_index - 1) \
                    * (map_object.tile_size[1] * scaling) + my_sprite.height / 2
                sprite_list.append(my_sprite)
    return sprite_list

def _create_sprite_from_tile(sprite_class, map_object, tile, scaling=1.0, base_directory=None):
    map_source = map_object.tmx_file
    map_directory = os.path.dirname(map_source)
    image_file = _get_image_source(tile, base_directory, map_directory)

    if tile.animation:
        raise NotImplementedError("Quest does not support animated tiles")
    else:
        image_x, image_y, width, height = _get_image_info_from_tileset(tile)

        my_sprite = sprite_class(image_file,
                           scaling,
                           image_x,
                           image_y,
                           width,
                           height)
        # Added by CP
        x1, y1 = - my_sprite._width / 2, - my_sprite._height / 2
        x2, y2 = + my_sprite._width / 2, - my_sprite._height / 2
        x3, y3 = + my_sprite._width / 2, + my_sprite._height / 2
        x4, y4 = - my_sprite._width / 2, + my_sprite._height / 2
        my_sprite._points = [[x1, y1], [x2, y2], [x3, y3], [x4, y4]]

    if tile.properties is not None and len(tile.properties) > 0:
        for my_property in tile.properties:
            my_sprite.properties[my_property.name] = my_property.value

    if tile.objectgroup is not None:

        if len(tile.objectgroup) > 1:
            print(f"Warning, only one hit box supported for tile with image {tile.image.source}.")

        for hitbox in tile.objectgroup:
            points: List[Point] = []
            if isinstance(hitbox, pytiled_parser.objects.RectangleObject):
                if hitbox.size is None:
                    print(f"Warning: Rectangle hitbox created for without a "
                          f"height or width for {tile.image.source}. Ignoring.")
                    continue

                sx = hitbox.location[0] - (my_sprite.width / (scaling * 2))
                sy = -(hitbox.location[1] - (my_sprite.height / (scaling * 2)))
                ex = (hitbox.location[0] + hitbox.size[0]) - (my_sprite.width / (scaling * 2))
                ey = -((hitbox.location[1] + hitbox.size[1]) - (my_sprite.height / (scaling * 2)))

                p1 = [sx, sy]
                p2 = [ex, sy]
                p3 = [ex, ey]
                p4 = [sx, ey]
                points = [p1, p2, p3, p4]

            elif isinstance(hitbox, pytiled_parser.objects.PolygonObject) \
                    or isinstance(hitbox, pytiled_parser.objects.PolylineObject):
                for point in hitbox.points:
                    adj_x = point[0] + hitbox.location[0] - my_sprite.width / (scaling * 2)
                    adj_y = -(point[1] + hitbox.location[1] - my_sprite.height / (scaling * 2))
                    adj_point = [adj_x, adj_y]
                    points.append(adj_point)

                if points[0][0] == points[-1][0] and points[0][1] == points[-1][1]:
                    points.pop()

            elif isinstance(hitbox, pytiled_parser.objects.ElipseObject):
                if hitbox.size is None:
                    print(f"Warning: Ellipse hitbox created for without a height "
                          f"or width for {tile.image.source}. Ignoring.")
                    continue

                hw = hitbox.size[0] / 2
                hh = hitbox.size[1] / 2
                cx = hitbox.location[0] + hw
                cy = hitbox.location[1] + hh

                acx = cx - (my_sprite.width / (scaling * 2))
                acy = cy - (my_sprite.height / (scaling * 2))
                total_steps = 8
                angles = [step / total_steps * 2 * math.pi for step in range(total_steps)]
                for angle in angles:
                    x = (hw * math.cos(angle) + acx)
                    y = (-(hh * math.sin(angle) + acy))
                    point = [x, y]
                    points.append(point)
            else:
                print(f"Warning: Hitbox type {type(hitbox)} not supported.")

            my_sprite.points = points

    return my_sprite



# Maze Walk Strategy

from quest.contrib.target_strategy import TargetStrategy
from quest.maze import Maze
from quest.map import GridMapLayer
from quest.helpers import Direction

class MazeWalkStrategy(TargetStrategy):
    """A strategy to randomly explore a maze.

    This Strategy relies on the insight that if you are stuck in a maze, if you keep your left 
    hand on the wall at all times and keep walking, you will eventually explore the whole maze.
    (There's nothing special about left; you could choose your right hand instead.) This assumes 
    a special definition of a maze, that there should be exactly one path between any two points. 
    No loops allowed. If you are not convinced, draw a few mazes and test it out. 

    To implement this insight, we can start with `quest.contrib.target_strategy.TargetStrategy`, which 
    knows how to direct a sprite toward a target until it gets there, and then to choose a new target.
    We just need to choose new targets in a way that gets the sprite around the maze.

    To choose a target, the sprite needs to have a `heading`, which keeps track of the direction
    it is currently facing. Then, if it's possible to turn left, always do that. Otherwise go straight. 
    If straight is not possible, then turn right. And if turning right is not possible either, then turn
    around and go back. (Update the heading with this new direction.)

    We need a SpriteList of walls so we can check if it's possible to go in a particular direction. 
    It's also helpful to have a grid--instead of using whatever random position the sprite is currently in,
    we can just think in grid squares. This simplifies the math a lot. 

    Arguments:
        grid (quest.map.GridMapLayer): The GridMapLayer used to build the maze.
        walls (arcade.SpriteList): A SpriteList containing all the walls.
    """

    def __init__(self, grid, walls):
        self.grid = grid
        self.wall_positions = [grid.get_grid_position((wall.center_x, wall.center_y)) for wall in walls]

    def choose_new_target(self, sprite, game):
        """Starting with left of the current heading (anticlockwise), keep turning 
        clockwise until a direction is not blocked. Set the sprite's heading to this 
        new direction and return the direction as a vector.

        Arguments:
            sprite: The sprite who is about to act.
            game: The game object (not used in this Strategy).
        """
        new_heading = sprite.heading.turn_anticlockwise()
        while self.direction_is_blocked(sprite, new_heading):
            new_heading = new_heading.turn_clockwise()
        sprite.heading = new_heading
        grid_target = self.get_new_grid_position(sprite, new_heading)
        sprite.target = self.grid.get_pixel_position(grid_target)
        
    def setup_sprite(self, sprite, game):
        """This strategy requires the sprite to remember the direction it is moving (its heading).
        If a sprite has no heading, we'll randomly assign one to get started.

        Arguments:
            sprite: The sprite who is about to act.
        """
        super().setup_sprite(sprite, game)
        if not hasattr(sprite, "heading"):
            sprite.heading = Direction.N

    def direction_is_blocked(self, sprite, direction):
        """Checks whether the direction (from the sprite's current position) is blocked by a wall.

        Arguments:
            sprite: The sprite who is about to act.
            direction (quest.helpers.Direction): The direction to check.
        """
        return self.get_new_grid_position(sprite, direction) in self.wall_positions

    def get_new_grid_position(self, sprite, direction):
        """First, we get the sprite's current position in grid coordinates (something like (3, 6), much
        nicer than pixel coordinates like (365.76, 190,88)). Then we calculate the new position by adding
        the direction's vector coordinates. (For example, Direction.N as a vector is (0, 1).)
        """

        current_x, current_y = self.grid.get_grid_position((sprite.center_x, sprite.center_y))
        direction_x, direction_y = direction.to_vector()
        return (current_x + direction_x, current_y + direction_y)



from quest.dialogue import Dialogue
from quest.modal import Modal, DialogueModal
from quest.examples.grandmas_soup import GrandmasSoupGame
from quest.helpers import resolve_resource_path

import os
from pathlib import Path

class GrandmasSoupVariablesGame(GrandmasSoupGame):
    """Help Grandma find all the ingredients for her soup.

    :py:class:`GrandmasSoupVariablesGame` extends the :py:class:`GrandmasSoupGame` to
    demonstrate the upgraded ink parser's support for variables.

        $ python -m quest.examples.grandmas_soup_variables

    After you play it, check out the sorce code by clicking on "source" in the
    blue bar just above.

    Attributes:
        dialogue: A :py:class:`Dialogue` containing the game's conversation.
        modal: A :py:class:`DialogueModal` to show the dialogue.
        dialogue_vars: A dictionary of dialogue variables which affect what is shown in the dialogue
    """

    def __init__(self):
        super().__init__()
        self.setup_dialogue()

    def setup_dialogue(self):
        """Sets up the dialogue with dialogue variables.
        """
        self.dialogue = Dialogue.from_ink(resolve_resource_path("grandma_variables.ink"))
        self.dialogue_vars =    {"num_veggies_left" : 4,
                                 "veggies" : "tomato, carrots, and potatoes"}
        self.dialogue.dialogue_vars = self.dialogue_vars
        self.modal = DialogueModal(self, self.dialogue)

    def got_item(self, description):
        """Adds the item's description to the items list. This is called when the
        player collides with a vegetable.

        Arguments:
            description: The item description.
        """
        self.items.append(description.upper())
        self.dialogue_vars["num_veggies_left"] -= 1    # decrement the variable tracking the number of veggies left for the dialogue
        if len(self.items) < 4:
            self.dialogue.run(self.items[-1])
        else:
            self.dialogue.run("SOUP")

if __name__ == '__main__':
    game = GrandmasSoupVariablesGame()
    game.run()



# timer.py
# by Xuxi9197
#
# Mixin to the physics engine to add a timer to the game.
# Update physics engine to call time_updates() in the update() function.

from time import time

class TimerMixin:
    """ Maintains a clock for the game that can enforce time-based game states
    and elements like a countdown.

    You can mix this class into a physics engine like this:
        class TimedPhysicsEngine(PhysicsEngine, TimerMixin):
        ...
        ...
    """
    def init_timer(self, total_game_time=None):
        """ Records the start time of the game

        Arguments:
            display (Boolean): whether to display the timer or not
            total_game_time (int): how much time a game should last
        """
        self.start_time = time()
        self.total_game_time = total_game_time

    def update(self):
        """ Updates the game based on the timer and stops the game from running if the total
        time of the game has elapsed.
        """
        
        if self.total_game_time:
            if self.time_since_start() >= self.total_game_time:
                self.game.running = False
        super().update()

    def time_since_start(self):
        """ Returns time since the start of the game.
        """
        return time() - self.start_time

    def time_remaining(self):
        """ Returns the time remaining or None if no total_game_time was set.
        """
        if self.total_game_time:
            return self.total_game_time - self.time_since_start()
        return None




import arcade
from quest.errors import NoLayerError, MultipleLayersError
from quest.sprite import QuestSprite
from quest.shim import process_layer
from math import floor

RED = 0
GREEN = 0
BLUE = 0

def clamp(bounds, val):
    low, high = bounds
    return max(low, min(high, val))

class Map:
    """Implements a map for a level or stage in the game.

    Each QuestGame may have multiple Maps. Each Map represents a game map with 
    multiple layers, allowing the map to describe walls, loot, and multiple layers
    of background imagery.

    Attributes:
        background_color: a 3-tuple of integers for red, green, blue. Each from 0-255.
            The :py:mod:`arcade:color` module also predefines many colors.
        tile_scaling: Factor by which to scale all map tiles. Default is 1.
    """
    background_color = (RED, GREEN, BLUE)
    tile_scaling = 1

    def __init__(self):
        """Initialize a Map with an empty layers list.
        """
        self.layers = []

    def add_layer(self, layer):
        """Add a layer the the layers list.

        Checks to make sure the layer's name is unique.

        Args:
            layer: The MapLayer to add.
        """
        for existing_layer in self.layers:
            if existing_layer.name == layer.name:
                raise ValueError("Map already has a layer named {}".format(layer.name))
        self.layers.append(layer)

    def get_layer_by_name(self, layer_name):
        """Looks up a map layer by name. 

        There should only be one layer with each name. If there is not exactly one layer,
        raises an error. 

        Args:
            layer_name: The name of the layer.

        Returns: The layer.
        """
        layers = [layer for layer in self.layers if layer.name == layer_name]
        if len(layers) == 0:
            raise NoLayerError("Map has no layer named {} (layers are: {})".format(layer_name, 
                    ", ".join(layer.name for layer in self.layers)))
        elif len(layers) > 1: 
            raise MultipleLayersError("Map has more than one layer named {}".format(layer_name))
        else:
            return layers[0]

class TiledMap(Map):
    """A subclass of Map which loads its layers from a TMX file. Each layer has a collection of tiles. 
    When TiledMap is initialized, a sprite is created for each of these tiles. The sprite's image and
    position are read from the tile layer data. The sprite's class (which can determine its behavior)
    can be set using the `sprite_classes` argument. 
    
    Use TiledMap when you want to design your map using
    [Tiled](https://www.mapeditor.org/). This app saves maps as TMX files.

    Arguments:
        filename (str): Path to the .tmx tilemap file
        sprite_classes: {layer_name: SpriteClass} dict specifying the sprite class
            that should be used for each layer. 
    """
    def __init__(self, filename, sprite_classes=None):
        super().__init__()
        tilemap = arcade.tilemap.read_tmx(filename)
        for layer_name, sprite_class in sprite_classes.items():
            sprite_list = process_layer(sprite_class, tilemap, layer_name, self.tile_scaling)
            layer = MapLayer(layer_name, sprite_list)
            self.add_layer(layer)

class MapLayer:
    """
    Each Map is made up of one or more MapLayers. 

    Arguments:
        name (str): The layer name. 
        sprite_list: An optional :py:class:`arcade.SpriteList`.
    """
    def __init__(self, name, sprite_list=None):
        self.name = name
        self.sprite_list = sprite_list or arcade.SpriteList()

    def draw(self):
        """Renders the layer by calling 
        """
        self.sprite_list.draw()

    def clear(self):
        """Delete all this layer's sprites.
        """
        while len(self.sprite_list) > 0:
            self.sprite_list.pop()

class GridMapLayer(MapLayer):
    """
    A MapLayer which is aware of grid coordinates. GridMapLayers are useful in situations when you 
    want to place sprites programmatically. Several methods are provided to support calculations
    about sprite positions with respect to the grid. This cam be simpler than working in pixel positions. 
    For example, the maze example program generates a maze
    and then uses a GridMapLayer to place the walls of the maze. 
    
    Arguments:
        name (str): Layer name.
        columns (int): Number of columns in the grid.
        rows (int): Number of rows in the grid.
        pixel_width (int): Width of the grid in pixels.
        pixel_height (int): Height of the grid in pixels.
        sprite_filename (str): Path to sprite image file. Only needed if you will be creating sprites
            on this grid layer.
        sprite_class: Class of sprites to create on this layer.

    """

    sprite_class = QuestSprite

    def __init__(self, name, columns, rows, pixel_width, pixel_height, sprite_filename=None, sprite_class=None):
        super().__init__(name)
        self.columns = columns
        self.rows = rows
        self.pixel_width = pixel_width
        self.pixel_height = pixel_height
        self.sprite_filename = sprite_filename
        if sprite_class:
            self.sprite_class = sprite_class

    def add_sprite(self, grid_x, grid_y, sprite=None):
        """Adds a sprite at a given grid position.

        Arguments:
            grid_x (int): The x-coordinate of the grid position 
            grid_y (int): The y-coordinate of the grid position 
            sprite (QuestSprite): (Optional) sprite to add to this layer and place at this grid position. If no
                sprite is given, a new sprite will be created.
        """
        sprite = sprite or self.create_sprite()
        sprite.left, sprite.bottom = self.get_pixel_position((grid_x, grid_y))
        self.sprite_list.append(sprite)

    def create_sprite(self):
        """Creates a sprite with image `self.sprite_filename` and class `self.sprite_class`. 
        """
        if self.sprite_filename is None:
            raise ValueError("Can't add sprites to GridMapLayer unless sprite_filename is defined.")
        return self.sprite_class(self.sprite_filename)

    def get_pixel_position(self, grid_position, center=True):
        """Converts pixel coordinates to grid coordinates.

        Arguments:
            grid_position (int, int): x and y grid coordinates 
            center (bool): By default, returns the pixel position of the center
            of the grid tile. When False, returns the pixel position of the lower
            left corner of the grid tile.

        Returns:
            (float, float) pixel position.
        """
        grid_x, grid_y = grid_position
        pixel_x = self.pixel_width * (grid_x / self.columns)
        pixel_y = self.pixel_height * (grid_y / self.rows)
        if center:
            pixel_x += (self.pixel_width / self.columns) / 2
            pixel_y += (self.pixel_height / self.rows) / 2
        return pixel_x, pixel_y

    def get_grid_position(self, pixel_position):
        """Converts grid position to pixel position. 

        Any pixel position within a grid tile will be converted to that
        grid tile's coordinates.

        Arguments: 
            pixel_position (float, float): x and y pixel coordinates.

        Returns:
            (float, float) grid position.
        """
        pixel_x, pixel_y = pixel_position
        grid_x = pixel_x // (self.pixel_width / self.columns)
        grid_y = pixel_y // (self.pixel_height / self.rows)
        return grid_x, grid_y

    def position_in_grid(self, grid_position):
        """Returns whether or not `grid_position` is within the grid.

        Arguments:
            grid_position (int, int): x and y grid coordinates 
        """
        grid_x, grid_y = grid_position
        return grid_x >= 0 and grid_x < self.columns and grid_y >= 0 and grid_y < self.rows



# island_time.py
#
# game that extends the basic island game to implement a clock that changes the
# appearance of the sprites based on the time of the game

from quest.examples.island import IslandAdventure
from quest.engines import ContinuousPhysicsEngine
from quest.contrib.timer import TimerMixin
import arcade
import os
from pathlib import Path

class TimedContinuousPhysicsEngine(ContinuousPhysicsEngine, TimerMixin):
    """ Physics engine that also makes time-based updates
    """

    def __init__(self, game):
        self.init_timer()
        super().__init__(game)

    def update(self):
        """Calculates shade by converting the time passed since the start of the game to a percentage
        of the current day/night cycle.
        """
        super().update()
        time_since_start = self.time_since_start() 
        curr_mod = time_since_start%self.game.time_cycle_secs
        grade = abs(curr_mod - self.game.time_cycle_secs/2) / (self.game.time_cycle_secs/2)
        color_value = grade*(255-self.game.max_darkness) + self.game.max_darkness
        for sprite in self.all_sprites:
            sprite.color = (color_value, color_value, color_value)


class IslandTime(IslandAdventure):
    """ A game that implements a day/night cycle by tracking the time since the game started
    and shading the sprites based on a day/night cycle
    """

    max_darkness = 100
    time_cycle_secs = 30
    display_time = True
    
    def __init__(self):
        super().__init__()
        

    def setup_physics_engine(self):
        """Passes optional `time` parameter to the standard :py:class:`ContinuousPhysicsEngine`.
        The result is that the sprites in the game change their appearance as time progresses in
        the game.
        """
        self.physics_engine = TimedContinuousPhysicsEngine(self)

    def message(self):
        """ Displays the time in the game relative to the time cycle
        """
        if self.display_time:
            return "Time: {}".format(int(self.physics_engine.time_since_start()%self.time_cycle_secs))

if __name__ == '__main__':
    game = IslandTime()
    game.run()



from quest.strategy import Strategy
from math import sqrt

def distance(a, b):
    """A helper function to find the distance between two points. 
    Uses the pythagorean theorem!

    Arguments:
        a: (x, y) point
        b: (x, y) point
    """
    ax, ay = a
    bx, by = b
    x = ax - bx
    y = ay - by
    return sqrt(x * x + y * y)

class TargetStrategy(Strategy):
    """A strategy to go toward a target until the target is reached
    (or another condition is satisfied). Then re-choose the target.

    Attributes:
        epsilon: A distance to consider "close enough," used to decide if the sprite reached
                 its target. (If you insist on exactly reaching the target, the sprite will
                 always overshoot a little, causing it to jiggle in place forever.)
    """

    epsilon = 10
    
    def choose_course(self, sprite, game):
        """Returns a (x, y) vector representing the chosen course.
        
        First, makes sure the sprite has a target. Then checks to see if a 
        new target is needed. If so, chooses a new target. If not, sets the course
        toward the target.
        """
        self.setup_sprite(sprite, game)
        if self.needs_new_target(sprite, game):
            self.choose_new_target(sprite, game)
        return self.course_to_target(sprite, game)

    def setup_sprite(self, sprite, game):
        """Makes sure the sprite has a target set.

        Arguments:
            sprite: The sprite who is about to act.
        """
        if not hasattr(sprite, "target"):
            sprite.target = (sprite.center_x, sprite.center_y)

    def needs_new_target(self, sprite, game):
        """Decides whether the sprite needs a new target. By default, 
        a sprite needs a new target if it has (nearly) reached its original target.
        """
        current_position = (sprite.center_x, sprite.center_y)
        return distance(current_position, sprite.target) <= self.epsilon

    def choose_new_target(self, sprite, game):
        """Chooses a new target and applies it to the sprite. By defualt, this does nothing.
        Subclasses of TargetStrategy should re-implement this method for desired behavior. 
        """
        sprite.target = sprite.target

    def course_to_target(self, sprite, game):
        """Returns a (x, y) vector representing the direction to the target.
        """
        target_x, target_y = sprite.target
        return (target_x - sprite.center_x, target_y - sprite.center_y)

    






# timer.py example game
# by Xuxi917 and Jacob Wolf
#
# Example game to demoinstrate mixing in a timer to a game

from quest.contrib.timer import TimerMixin
from quest.examples.maze import MazeGame
from quest.engines import ContinuousPhysicsEngine

TIME_LIMIT = 30

class TimedContPhysicsEngine(TimerMixin, ContinuousPhysicsEngine):
    """ Physics engine that also makes time-based updates
    """

    def __init__(self, game):
        """ Starts timer.
        """
        self.init_timer(TIME_LIMIT)
        super().__init__(game)

class TimedMazeGame(MazeGame):
    """ Generates a game with MazeGame with a time limit using the TimerMixin.
    """

    def setup_physics_engine(self):
        """ Sets up a physics engine with a timer using the TimerMixin
        """
        self.physics_engine = TimedContPhysicsEngine(self)

    def message(self):
        """Returns a string representation of the time remaining in the game.
        """
        if self.physics_engine.time_remaining() <= 0:
            return "You got {} stars!".format(self.score)
        else:
            return "Time remaining: {}".format(int(self.physics_engine.time_remaining()))

if __name__ == '__main__':
    game = TimedMazeGame()
    game.run()





from random import random
from math import tau, sin, cos

class Strategy:
    """A Strategy represents how a sprite responds to the environment. 

    It is useful to extract :py:class:`Strategy` as a separate class (rather than
    making it part of the NPC class) because there might be many sprites in a game
    which need to use the same strategy. In the default implementation of 
    :py:meth:`quest.sprite.NPC.on_update`, the NPC will set its course using its 
    :py:class:`Strategy` if it has one. 
    """

    def choose_course(self, sprite, game):
        """Returns a (x, y) vector representing the chosen course.

        In the default implementation, returns (0, 0) which will cause the sprite to
        stop. To create new strategies, override this method and define helper methods
        as necessary.

        Arguments:
            sprite: The sprite who is about to act.
            game: The game object (to access attributes useful in choosing the course).
        """
        return (0, 0)

class RandomWalk(Strategy):
    """ A strategy which causes the sprite to walk randomly, with some chance 
    of changing direction at each step.

    Arguments:
        change_prob (float): The probability (between 0 and 1) that the sprite will choose
        a new direction on a step.
    """
    def __init__(self, change_prob=0.1):
        self.change_prob = change_prob
        self.set_random_direction()

    def choose_course(self, sprite, game):
        """Possibly chooses a new random direction, then returns the current direction.

        Arguments:
            sprite: The sprite who is about to act.
            game: The game object (to access attributes useful in choosing the course).
        """
        if random() < self.change_prob:
            self.set_random_direction()
        return (self.x, self.y)

    def set_random_direction(self):
        """Sets attributes `x` and `y` to a random (x, y) direction. 

        Think of this as choosing a random point on a circle. Tau is another 
        word for 2 * pi, which represents the angle measure of the full circle, so 
        if we multiply it by `random()` (which returns a random number evenly distributed 
        between 0 and 1), we get a randomly-chosen angle measure between 0 and the full circle. 
        Then we can use cosine and sine to get the x and y components of this point.
        """
        r = random() * tau
        self.x = cos(r)
        self.y = sin(r)

class DividedStrategy(Strategy):
    """Builds a complex strategy based on two existing strategies and transition probabilities.

    One way to get more complex behavior is to define two simpler behaviors, and then to have a 
    certain probability of switching between them. For example, some games have enemy NPCs which 
    sometimes chase the player and other times wander around randomly. The transition probabilities
    (the chance of switching) are not symmetric. For example, you might want to define a game where 
    sprites have a low chance of going from normal to desperate, and then zero chance of going from
    desperate back to normal. The idea of composing strategies from other strategies is an example of 
    functional programming.

    This pattern could be extended to `n` strategies, with a matrix of `n * n` transition probabilities.
    This more general case is called a Markov chain, and gets used quite a lot in computer science. 

    Arguments:
        strategy_a (:py:class:`Strategy`): The initial strategy.
        strategy_b (:py:class:`Strategy`): A second strategy.
        ab_prob (float): The probability of switching from strategy a to strategy b.
        ba_prob (float): The probability of switching from strategy b to strategy a.
    """

    def __init__(self, strategy_a, strategy_b, ab_prob, ba_prob):
        self.strategy_a = self.current_strategy = strategy_a
        self.strategy_b = strategy_b
        self.ab_prob = ab_prob
        self.ba_prob = ba_prob

    def choose_course(self, sprite, game):
        """Possibly switch strategies, and then have the current strategy handle choosing the course.

        Arguments:
            sprite: The sprite who is about to act.
            game: The game object (to access attributes useful in choosing the course).
        """
        self.consider_switching_strategy()
        return self.current_strategy.choose_course(sprite, game)

    def consider_switching_strategy(self):
        """Possibly switch strategy (using transition probabilities).
        """
        if self.current_strategy is self.strategy_a:
             if random() < self.ab_prob:
                self.current_strategy = self.strategy_b
        else:
            if random() < self.ba_prob:
                self.current_strategy = self.strategy_a



class MouseMotionMixin:
    """Extends `quest.QuestGame` to allow the player to be controlled by the mouse.
    Example usage::

        from quest.examples.island import IslandAdventure
        from quest.contrib.mouse import MouseMotionMixin

        class IslandWithMouse(MouseMotionMixin, IslandAdventure):
            "The island adventure with mouse control."

        IslandWithMouse().run()
    """

    def on_mouse_motion(self, x: float, y: float, dx: float, dy: float):
        """
        Keep track of the mouse's new coordinates whenever it moves.
        This method is called automatically whenever the mouse moves.

        :param float x: x position of mouse
        :param float y: y position of mouse
        :param float dx: Change in x since the last time this method was called
        :param float dy: Change in y since the last time this method was called
        """

        if self.player.center_y != y:
            self.player.change_y = dy
        if self.player.center_x != x:
            self.player.change_x = dx



from quest.examples.island import IslandAdventure
from quest.sprite import Player
from quest.contrib.hit_points import HitPointsMixin
import arcade

class MortalPlayer(HitPointsMixin, Player):
    """MortalPlayer uses `HitPointsMixin` to add hit point
    properties and methods like `change_hit_points`. Here, we re-define
    `on_collision` to lower the player's hit points whenever she bumps
    into another sprite (like a wall). Because the default behavior of
    `on_min_hit_points` is to kill the sprite, the player will be removed
    when she runs out of hit points.
    """

    max_hit_points = 3
    hit_points = 3

    def on_collision(self, other_sprite, game):
        self.change_hit_points(-1)
    

class DangerousIsland(IslandAdventure):
    """Extends `IslandAdventure` to use `MortalPlayer` and to 
    constantly report the player's hit points.
    """

    def setup_player(self):
        """Creates the player sprite. Almost the same as the regular 
        `setup_player`, except uses `MortalPlayer` instead of `Player`.
        """
        self.player = MortalPlayer(self.player_sprite_image, self.player_scaling)
        self.player.center_x = self.player_initial_x
        self.player.center_y = self.player_initial_y
        self.player.speed = self.player_speed
        self.player_list = arcade.SpriteList()
        self.player_list.append(self.player)

    def message(self):
        return "Hit points: {}/{}".format(
            self.player.hit_points,
            self.player.max_hit_points,
        )
        
if __name__ == '__main__':
    game = DangerousIsland()
    game.run()



from quest.game import QuestGame
from quest.map import Map, GridMapLayer, TiledMap
from quest.sprite import Background, Wall
from quest.engines import DiscretePhysicsEngine
from quest.helpers import resolve_resource_path
import os
from pathlib import Path

class IslandAdventureDiscrete(QuestGame):
    """An alternate version of Island Adventure, using discrete movement.

    To run this example::

        $ python -m quest.examples.island_discrete

    """
    player_sprite_image = resolve_resource_path("images/boy_simple.png")
    player_initial_x = 300
    player_initial_y = 300
    player_speed = 6

    def setup_maps(self):
        """Sets up the map.

        Uses a :py:class:`TiledMap` to load the map from a ``.tmx`` file,
        created using :doc:`Tiled <tiled:manual/introduction>`.
        """
        super().setup_maps()
        sprite_classes = {
            "Obstacles": Wall,
            "Background": Background,
        }
        self.add_map(TiledMap(resolve_resource_path("images/island/island.tmx"), sprite_classes))
        layer = GridMapLayer("grid", 40, 40, 40*32, 40*32)
        self.get_current_map().add_layer(layer)

    def setup_walls(self):
        """Assigns sprites to `self.wall_list`. These sprites will function as walls, blocking
        the player from passing through them.
        """
        self.wall_list = self.get_current_map().get_layer_by_name("Obstacles").sprite_list

    def setup_physics_engine(self):
        """Uses :py:class:`DiscretePhysicsEngine` instead of the standard :py:class:`ContinuousPhysicsEngine`.
        The result is that the player snaps to a grid instead of moving smoothly to any position.

        A game's physics engine is responsible for enforcing the rules of the game's reality.
        In a fancy 3d game, the physics engine is full of intense math to keep track of objects
        flying around, bouncing off of walls, and breaking into pieces.

        Quest's physics engines are simpler. They just need to make sure nobody walks through walls,
        and to check when sprites collide. There are two physics engines built-in: A continuous (the
        default) and a discrete (used here).
        """
        grid = self.get_current_map().get_layer_by_name('grid')
        self.physics_engine = DiscretePhysicsEngine(self, grid)

if __name__ == '__main__':
    game = IslandAdventureDiscrete()
    game.run()



from collections import defaultdict
import random

def is_even(x):
    "Returns True if x is even, otherwise False."
    return x % 2 == 0

def is_odd(x):
    "Returns True if x is odd, otherwise False."
    return not is_even(x)

class Maze:
    """Generates a maze stored in a 2-dimensional array.

    This class generates a maze. The mathematical definition of a maze is a set 
    of paths that are connected so that there is exactly one way to go between 
    any two points. Here's an example::

        >>> from quest.maze import Maze
        >>> m = Maze(55, 15)
        >>> m.generate()
        >>> print(m)

        ╔═══════════════╦═════╦═════════╦═════════╦═══════╦═══╗
        ║               ║     ║         ║         ║       ║   ║
        ║ ╔═════════╦══ ║ ══╗ ║ ║ ══╦══ ╚═══╦══ ║ ║ ╔═══╗ ╠══ ║
        ║ ║         ║   ║   ║   ║   ║       ║   ║   ║   ║ ║   ║
        ║ ║ ╔═════╗ ║ ║ ╚═╗ ╚═╦═╩═╗ ╚═╦═══╗ ║ ╔═╩═══╝ ║ ║ ║ ══╣
        ║ ║ ║     ║ ║ ║   ║   ║   ║   ║   ║   ║       ║ ║ ║   ║
        ║ ║ ║ ╔═╗ ║ ║ ╠═══╝ ╔═╝ ║ ║ ║ ║ ║ ╚═══╣ ══╗ ╔═╩═╝ ╠══ ║
        ║ ║ ║ ║ ║ ║ ║ ║     ║   ║ ║ ║ ║ ║     ║   ║ ║     ║   ║
        ║ ║ ║ ║ ║ ║ ║ ║ ══╦═╝ ╔═╝ ╠═╝ ║ ╚═══╗ ╠══ ║ ║ ╔═══╣ ║ ║
        ║ ║     ║ ║ ║     ║   ║   ║   ║     ║ ║   ║   ║   ║ ║ ║
        ╠═╩═════╝ ║ ╠═══╦═╝ ╔═╣ ╔═╝ ╔═╩═╦══ ║ ║ ╔═╩═══╝ ║ ║ ╚═╣
        ║         ║ ║   ║   ║ ║ ║   ║   ║   ║   ║       ║ ║   ║
        ║ ╔═══════╝ ║ ║ ║ ╔═╝ ║ ╚══ ║ ══╝ ══╩═══╩════ ╔═╝ ╚══ ║
        ║ ║           ║   ║         ║                 ║       ║
        ╚═╩═══════════╩═══╩═════════╩═════════════════╩═══════╝

    There are many different algorithms for generating mazes, and they tend to 
    produce mazes with different characteristics. This class implements a 
    `depth-first search <http://algostructure.com/specials/maze.php>`_ maze 
    generation algorithm, which tends to produce long corridors without much
    branching. See :py:meth:`generate` for a description of how this works.

    Args:
        columns: the number of columns in the maze (including edge walls).
        rows: the number of rows in the maze (including edge walls).

    .. _depth-first search: http://algostructure.com/specials/maze.php
    """

    def __init__(self, columns, rows):
        self.columns = columns
        self.rows = rows

    def generate(self, seed=None):
        """Generates (or re-generates) a random maze.

        We are always going to keep track of a current point and we're going
        to keep a list of points that have been visited. Every time a point 
        becomes the current point, we will add it to `visited`. Also, every
        time we move to a new current point, we add the old current point to a 
        stack. This becomes a history of where we have been. 

        Now we see if the current point has any neighbors that have not yet been 
        visited. If so, choose one randomly, make it the current point, and 
        repeat. Now the maze is growing like a worm, one point at a time. At 
        some point, though, the growing worm will get stuck in a dead end, where
        the current point has no unvisited neighbors. It can't grow anymore.

        This is where the history stack comes in. Since the maze can't grow from
        the current point, let's pop the most recent point off the history stack
        and make that the current point instead. If that point has unvisited
        neighbors, great: we can continue. Otherwise, keep popping points off
        the history stack until we find one that has unvisited neighbors. 

        When does this end? Once the history stack is empty. That means we have
        worked our way all the way back to the starting point, and no points 
        have any unvisited neighbors remaining. This means the maze has filled
        up its whole world so it's finished!

        If you want to see this in action, 
        `here <http://algostructure.com/specials/maze.php>`_ is a website with 
        several different maze-generation algorithms. Choose 
        "Flood fill/Depth-first."

        Args:
            seed: If provided, sets the random seed. (Random numbers on 
                computers are not really random, they're just hard to predict. If
                you set the random seed to the same value, you'll get the same set
                of random numbers every time. This is helpful when you want to get
                the same random maze.)
        """
        if seed:
            random.seed(seed)
        self.links = defaultdict(set)
        visited = set()
        stack = [(1, 1)]
        current_point = (1, 1)
        while len(stack) > 0:
            visited.add(current_point)
            unvisited_neighbors = [n for n in self.neighbors(current_point) if n not in visited]
            if len(unvisited_neighbors) > 0:
                next_point = random.choice(unvisited_neighbors)
                self.connect(current_point, next_point)
                stack.append(current_point)
                current_point = next_point
            else:
                current_point = stack.pop()

    def generate_fully_connected_maze(self):
        """Generates a maze where every node is connected to all its neighbors.
        """
        self.links = defaultdict(set)
        x = 1
        while x < self.columns:
            y = 1
            while y < self.rows:
                for n in self.neighbors((x, y)):
                    self.links[(x, y)].add(n)
                y += 2
            x += 2

    def connect(self, p0, p1):
        """Connects two points by storing each in the other's list of links.

        Args:
            p0: A point, which is represented as a tuple of (x, y) coordinates.
            p1: Another point, also a tuple.
        """
        self.links[p0].add(p1)
        self.links[p1].add(p0)
                
    def connected(self, p0, p1):
        """Checks whether two points are connected as neighbors.

        Args:
            p0: A point, which is represented as a tuple of (x, y) coordinates.
            p1: Another point, also a tuple.

        Returns:
            True if the points are connected, otherwise False.
        """
        return p1 in self.links[p0]

    def neighbors(self, point):
        """Gets a list of the point's neighbors.

        The points we care about are those with odd coordinates,
        like (1, 1) and (5, 7). Let's call these points nodes. They 
        are represented with stars in the diagrams below. 

        The reason nodes are on the odd coordinates is because we need
        to leave room for walls between points in the maze, and surrounding
        the maze. 
        ::

            ╔═══════╗   ╔═══════╗   ╔═══════╗
            ║* * * *║   ║* * * *║   ║       ║
            ║ ┼ ┼ ┼ ║   ╠═════╗ ║   ╠═════╗ ║
            ║* * * *║   ║* * *║*║   ║     ║ ║
            ║ ┼ ┼ ┼ ║   ║ ╔══ ║ ║   ║ ╔══ ║ ║
            ║* * * *║   ║*║* *║*║   ║ ║   ║ ║
            ║ ┼ ┼ ┼ ║   ║ ║ ══╝ ║   ║ ║ ══╝ ║
            ║* * * *║   ║*║* * *║   ║ ║     ║
            ╚═══════╝   ╚═╩═════╝   ╚═╩═════╝

        Args:
            point: A tuple of (x, y) coordinates.

        Returns:
            A list of nodes two spaces to the left, right, down, and up from 
            the given point, as long as they are in bounds.
        """
        x, y = point
        possible_neighbors = [(x+2, y), (x-2, y), (x, y+2), (x, y-2)]
        return [n for n in possible_neighbors if self.is_in_bounds(n)]

    def is_in_bounds(self, point):
        """Checks whether a point is in bounds.

        Returns:
            True if the point is in bounds, otherwise False.
        """
        x, y = point
        return x >= 0 and x < self.columns - 1 and y >= 0 and y < self.rows - 1

    def get_walls(self):
        """Returns a list of points containing walls in the current maze.
        
        To find all the walls, we just loop through every (x, y) point in the 
        maze and check if it's a wall. First off, all the points around the 
        edges must be walls. Then, If both a point's x and y coordinates are 
        odd, the point is definitely not a wall (See :py:meth:`neighbors`).
        If both a point's coordinates are even, it is definitely a wall.
        For points with one even and one odd coordinate, we need to check the
        links to see whether the point has been chosen as a link. Otherwise, 
        it's a wall.

        Returns:
            A list of (x, y) tuples for wall spaces in the maze.
        """
        walls = []
        for i in range(self.columns):
            for j in range(self.rows):
                if i == 0 or i == self.columns - 1 or j == 0 or j == self.rows - 1:
                    walls.append((i, j))
                elif is_even(i) and is_even(j):
                    walls.append((i, j))
                elif is_even(i) and (i+1, j) not in self.links[(i-1, j)]:
                    walls.append((i, j))
                elif is_even(j) and (i, j+1) not in self.links[(i, j-1)]:
                    walls.append((i, j))
        return walls

    def __str__(self, nodes=False):
        """Produces a string representation of the maze (example above).

        Since the maze is stored in an array of arrays, we can just turn each 
        array (row of the maze) into a string and then join the rows together 
        with newline characters.

        Args:
            nodes: If True, points that are nodes (see :py:meth:`neighbors`)
                will be shown as stars.

        Returns:
            A multiline string representation of the maze.
        """
        walls = set(self.get_walls())
        rows = []
        for j in reversed(range(self.rows)):
            row = [self.str_char((i, j), walls, nodes) for i in range(self.columns)]
            rows.append(''.join(row))
        return "\n".join(rows)

    def str_char(self, point, walls, nodes=False):
        """Determines which character should represent a point in the maze.

        Args:
            point: A (x, y) tuple.
            walls: A list of (x, y) tuples.
            nodes: If True, points that are nodes (see :py:meth:`neighbors`)
                will be shown as stars.
        """
        x, y = point
        symbols = " ═║╚══╝╩║╔║╠╗╦╣╬"
        code = 0
        if nodes and is_odd(x) and is_odd(y):
            return '*'
        if (x, y) in walls:
            if (x+1, y) in walls:
                code += 1
            if (x, y+1) in walls:
                code += 2
            if (x-1, y) in walls:
                code += 4
            if (x, y-1) in walls:
                code += 8
        return symbols[code]

if __name__ == '__main__':
    m = Maze(99, 99)
    m.generate()
    print(m)



# npchase.py
# by NpChase and Jacob Wolf

from quest.strategy import Strategy

class Chase(Strategy):
    def choose_course(self, sprite, game):
        """Moves the NPC towards the character.

        Arguments:
            sprite: The sprite who is about to act.
            game: The game object (to access attributes useful in choosing the course).

        Returns:
            (int, int) pair of the amount to change in the x and y direction
        """
        x_npc = sprite.center_x
        y_npc = sprite.center_y
        x_player = game.player.center_x
        y_player = game.player.center_y
        x_int = x_player - x_npc
        y_int = y_player -  y_npc
        return (x_int, y_int)



class HitPointsMixin:
    """A mixin for `quest.Sprite` which adds hit points to a sprite.

    Attributes:
        max_hit_points (int): hit_points can't go above this value. Default 100.
        min_hit_points (int): hit_points can't go below this value. Default 0.
        hit_points (int): the sprite's current hit points
    """

    max_hit_points = 100
    min_hit_points = 0
    hit_points = 100

    def change_hit_points(self, delta):
        """Changes hit points by `delta`, keeping hit points within bounds. 
        Calls on_max_hit_points and on_min_hit_points if needed.
        """
    
        self.hit_points += delta
        if self.hit_points >= self.max_hit_points:
            self.hit_points = self.max_hit_points
            self.on_max_hit_points()
        if self.hit_points <= self.min_hit_points:
            self.hit_points = self.min_hit_points
            self.on_min_hit_points()

    def on_max_hit_points(self):
        """Called whenever the sprite's hit points reach the maximum
        By default, does nothing.
        """
        pass

    def on_min_hit_points(self):
        """Called whenever the sprite's hit points reach the maximum
        By default, kills the sprite.
        """
        self.kill()



from quest.helpers import SimpleInkParser

class Dialogue:
    """Models a dialogue interaction as a graph of knots. Each knot has a list of content
    and then a list of choices leading to further knots (or ending the dialogue).
    where each choice links to another talk turn or ends the dialogue. A Dialogue
    instance has a "result" property which the game can use to determine what
    happened.
    """

    @classmethod
    def from_ink(cls, ink_path):
        """Reads talk turns from a YAML file and uses them to initialize a new
        Dialogue. Note that this is a @classmethod, which means it should be used
        by the Dialogue class, not an instance of Dialgoue. For example:

            >>> dialogue = Dialogue.from_ink("examples/dialogue_sample.ink")

        Arguments:
            ink_path: A string with the ink filename containing talk turns.
                talk turns. For an example of the expected format, see
                `quest/examples/dialogue_sample.ink`.

        Returns:
            A newly-created Dialogue instance.
        """
        parser = SimpleInkParser()
        with open(ink_path) as dialogue_file:
            knots = parser.parse(dialogue_file)
        return Dialogue(knots)

    def __init__(self, knots):
        self.knots = knots
        self.dialogue_vars = None
        self.run()

    def run(self, start_at="START"):
        """Starts (or re-starts) a run of the dialogue by setting the
        current knot to `start_at` and `running` to True.

        Arguments:
            start_at: (default "START") Name of knot to start at.
        """
        self.current_knot = start_at
        self.knots_visited = [self.current_knot]
        self.running = True

    def get_content(self):
        """Returns a list of the content for the current knot.
        """
        content =  self.knots[self.current_knot]['content']
        if self.dialogue_vars:
            content = self.replace_vars(content)
        return content

    def get_options(self):
        """Returns a list of option text for the current knot.
        """
        options = list(self.knots[self.current_knot]['options'].keys())
        if self.dialogue_vars:
            options = self.replace_vars(options)
        return options

    def choose(self, choice):
        """Chooses an option and moves to another knot. If that knot is "END",
        ends the current dialogue run.

        Arguments:
            choice (int): The index of the choice to follow.
        """
        choices = list(self.knots[self.current_knot]['options'].values())
        if choices[choice] == 'END':
            self.running = False
        else:
            self.current_knot = choices[choice]
            self.knots_visited.append(self.current_knot)

    def replace_vars(self, contents):
        """Returns the content items of a knot or option content list with the ink variables replaced with variables
        from the game's dialogue_vars dictionary.

        Arguments:
            content (str): the ink content
            dialogue_vars (dict): the variables from the game associated with their value
        """
        var_filled_contents = []
        for content_item in contents:
            var_filled_content_item = []
            content_item_vars = content_item.split("}")
            for num, text_with_var in enumerate(content_item_vars):
                if num == len(content_item_vars)-1:                  # last item in list, no variables, append whole string
                    var_filled_content_item.append(text_with_var)
                else:                                                # replace variables and append
                    if text_with_var[-1] == "\\":   # found "}" was actually "\}"
                        var_filled_content_item.append(text_with_var)
                        var_filled_content_item.append("}")
                    else:                           # replace ink variable
                        static_text, ink_var = text_with_var.rsplit("{", 1)
                        var_filled_content_item.append(static_text)
                        try:
                            ink_var_value = self.dialogue_vars[ink_var]
                        except Exception as e:
                            raise ValueError(f"Variable {ink_var} from ink file not found in dialogue_vars dictionary.")
                        var_filled_content_item.append(str(ink_var_value))
            content_item_string = "".join(var_filled_content_item)
            content_item_string = content_item_string.replace("\\{", "{")
            content_item_string = content_item_string.replace("\\}", "}")
            var_filled_contents.append(content_item_string)
        return var_filled_contents

class DialogueRunner:
    """Runs a Dialgoue from the command line. This is useful for testing out dialogue.
    If you have a file called "conversation.ink", here's an example of how to run it:

        >>> dialogue = Dialogue.from_ink("conversation.ink")
        >>> runner = DialogueRunner()
        >>> runner.run(dialogue)
    """
    def run(self, dialogue, start_at="START", dialogue_vars = None):
        dialogue.run(start_at=start_at)
        while dialogue.running:
            print('-' * 80)
            for content in dialogue.get_content(dialogue_vars):
                print(content + '\n')
            for i, option in enumerate(dialogue.get_options(dialogue_vars)):
                print("{}) {}".format(i, option))
            dialogue.choose(self.get_numeric_choice(maximum=i))

    def get_numeric_choice(self, maximum=None):
        choice = input("> ")
        while not choice.isdigit() and int(choice) <= maximum:
            print("Please enter a number between 0 and {}.".format(maximum))
            choice = input("> ")
        return int(choice)

if __name__ == '__main__':
    dialogue = Dialogue.from_ink("examples/dialogue_sample.ink")
    runner = DialogueRunner()
    runner.run(dialogue)



class PlayerEnabledMixin:
    """Extends `quest.QuestGame` to allow setting whether the player is enabled or not.
    When the player is disabled, user input is ignored. Note that the game may still take 
    actions affecting the player sprite, just not the user.

    Attrs:
        player_enabled (bool): When True (default), user input controls the player.

    """

    player_enabled = True

    def on_key_press(self, key, modifiers):
        """Delegates to the regular `on_key_press` method when `player_enabled` is True.

        Arguments:
            key: The key that was pressed.
            modifiers: A list of currently-active modifier keys (e.g. shift).
        """
        if self.player_enabled:
            super().on_key_press(key, modifiers)

    def on_mouse_motion(self, x: float, y: float, dx: float, dy: float):
        """
        Delegates to the regular `on_mouse_motion` when `player_enabled` is True.

        :param float x: x position of mouse
        :param float y: y position of mouse
        :param float dx: Change in x since the last time this method was called
        :param float dy: Change in y since the last time this method was called
        """
        if self.player_enabled:
            super().on_mouse_motion(x, y, dx, dy)



from quest.game import QuestGame
from quest.map import TiledMap
from quest.sprite import Background, Wall
from quest.helpers import resolve_resource_path
import arcade
import os
from pathlib import Path

class IslandAdventure(QuestGame):
    """A very simple subclass of :py:class:`QuestGame`.

    To run this example::

        $ python -m quest.examples.island

    :py:class:`IslandAdventure` shows off the basic features of the Quest 
    framework, loading a map and letting the player explore it. 
    After you play it, check out the sorce code by clicking on "source" in the
    blue bar just above.
    """

    player_sprite_image = resolve_resource_path("images/boy_simple.png")
    screen_width = 500
    screen_height = 500
    left_viewport_margin = 96                            
    right_viewport_margin = 96
    bottom_viewport_margin = 96
    top_viewport_margin = 96
    player_initial_x = 300
    player_initial_y = 300
    player_speed = 6

    def setup_maps(self):
        """Sets up the map.

        Uses a :py:class:`TiledMap` to load the map from a ``.tmx`` file,
        created using :doc:`Tiled <tiled:manual/introduction>`. 
        """
        super().setup_maps()
        sprite_classes = {
            "Obstacles": Wall,
            "Background": Background,
        }
        island_map = TiledMap(resolve_resource_path("images/island/island.tmx"), sprite_classes)
        self.add_map(island_map)

    def setup_walls(self):
        """Assigns sprites to `self.wall_list`. These sprites will function as walls, blocking
        the player from passing through them.
        """
        self.wall_list = self.get_current_map().get_layer_by_name("Obstacles").sprite_list

if __name__ == '__main__':
    game = IslandAdventure()
    game.run()



from quest.game import QuestGame
from quest.map import Map, GridMapLayer
from quest.maze import Maze
from quest.sprite import Wall, NPC
from quest.helpers import resolve_resource_path
from itertools import product
import arcade
import random
from datetime import datetime
import os
from pathlib import Path

class MazeGame(QuestGame):
    """Get all the stars as fast as you can! My record is 45 seconds.

    :py:class:`MazeGame` is an example of how you can make a fairly complex 
    game without making too many changes. Because :py:class:`MazeGame` is a 
    subclass of :py:class:`QuestGame`, we just need to change the parts we 
    want to work differently. We need to set some of the :py:class:`MazeGame` 
    properties and override a few of the class methods. 

    To run this example::

        $ python -m quest.examples.maze

    After you play it, check out the sorce code by clicking on "source" in the
    blue bar just above.


    Attributes:
        tile_size=32: Each square tile in the map is 32 pixels across.
        grid_columns=33: The map will have 33 columns of tiles.
        grid_rows=33: The map will have 33 rows of tiles.
        player_sprite_image="images/boy_simple.png": The sprite's image file.
        player_scaling=0.5: The image is too big, so we scale it down. (We could
            also just resize the image itself.)
        player_initial_x=1.5 * tile_size: By starting the player at 1.5 times `tile_size`, 
            the player will initially be positioned at the center of tile (1, 1).
            The outer edge of the map is walls.
        player_initial_y=1.5 * tile_size: Again, centering the player at (1, 1)
        score=0: Keeps track of how much loot has been collected.
        max_score=25: Total amount of loot to be distributed through the maze.
        game_over=False: Keeps track of whether the game has ended.
    """
    tile_size = 32
    grid_columns = 33
    grid_rows = 33
    player_sprite_image = resolve_resource_path("images/boy_simple.png")
    player_scaling = 0.5
    player_speed = 5
    player_initial_x = 1.5 * tile_size
    player_initial_y = 1.5 * tile_size
    score = 0
    max_score = 25
    game_over = False

    def __init__(self):
        super().__init__()
        self.level_start = datetime.now()

    def setup_maps(self):
        """Creates a :py:class:`MazeMap` (see below) and adds it to game's list of maps.
        """
        super().setup_maps()
        maze_map = MazeMap(self.grid_columns, self.grid_rows, self.tile_size, self.max_score)
        self.add_map(maze_map)

    def setup_walls(self):
        """Assigns `self.wall_list` to be all the sprites in the map's "walls" layer.
        """
        self.wall_list = self.get_current_map().get_layer_by_name("walls").sprite_list

    def setup_npcs(self):
        """Assigns `self.npc_list` to be all the sprites in the map's "loot" layer.
        """
        self.npc_list = self.get_current_map().get_layer_by_name("loot").sprite_list

    def on_loot_collected(self, collector):
        """A method to be called whenever loot is collected.

        See the :py:class:`Loot` NPC sprite below. It calls this method whenever there is
        a collision with the player.
        """
        self.score += 1
        if self.score == self.max_score:
            self.game_over = True
            self.final_elapsed_time = (datetime.now() - self.level_start).seconds

    def message(self):
        """Returns a string like "Time 12 (12/25)"
        """
        if self.game_over:
            return "You won in {} seconds!".format(self.final_elapsed_time)
        else:
            elapsed_time = datetime.now() - self.level_start
            return "Time {} ({}/{})".format(elapsed_time.seconds, self.score, self.max_score)

class Loot(NPC):
    """Loot is a NPC which shows up in the game as a star. Its only job is to
    get collected by the player.
    """
    def on_collision(self, sprite, game):
        """When the player collides with a Loot, it calls :py:meth:`quest.maze.MazeMap.on_loot_collected` to tell
        the game to make needed updates. Then the Loot kills itself.
        """
        game.on_loot_collected(sprite)
        print("Got a star!")
        self.kill()

class MazeMap(Map):
    """A Map which creates a wall layer using a :py:class:`Maze`.

    :py:class:`MazeMap` is a subclass of :py:class:`Map` which automatically generates a 
    maze. It uses a :py:class:`Maze` to figure out where to put walls, 
    and adds wall sprites to a map layer in a corresponding pattern.
    Args:
        columns: The number of columns of tiles in the map
        rows: The number of rows of tiles in the map
        tile_size: The size (in pixels) of each square tile
        num_loot: The number of loot sprites to add to the map
        
    """
    def __init__(self, columns, rows, tile_size, num_loot):
        super().__init__()
        self.columns = columns
        self.rows = rows
        self.tile_size = tile_size
        self.num_loot = num_loot
        self.maze = Maze(self.columns, self.rows)

        self.background_color = (20,80,20)
        self.add_layer(self.get_wall_map_layer())
        self.add_layer(self.get_loot_map_layer())
        self.generate_maze()

    def generate_maze(self, seed=None):
        """Generates (or re-generates) the map. The :py:class:`Maze` does most of the work.

        Regenerates the maze, clears the wall map layer and the loot map layer (in case
        there was a previous maze), and then populates these layers with new wall sprites
        and loot sprites.
        
        Args:
            seed: Random seed to pass to the maze (see :py:meth:`Maze.generate`)
        """
        self.maze.generate(seed)
        wall_map_layer = self.get_layer_by_name("walls")
        wall_map_layer.clear()
        for x, y in self.maze.get_walls():
            wall_map_layer.add_sprite(x, y)
        loot_map_layer = self.get_layer_by_name("loot")
        loot_map_layer.clear()
        for x, y in random.sample(self.possible_loot_locations(), self.num_loot):
            loot_map_layer.add_sprite(x, y)

    def get_wall_map_layer(self):
        """Creates a new :py:class:`GridMapLayer` to hold walls.
        """
        return GridMapLayer(
            name="walls",
            columns=self.columns,
            rows=self.rows,
            pixel_width=self.columns * self.tile_size,
            pixel_height=self.rows * self.tile_size,
            sprite_filename=resolve_resource_path("images/box.png"),
            sprite_class=Wall,
        )

    def get_loot_map_layer(self):
        """Creates a new :py:class:`GridMapLayer` to hold loot.
        """
        return GridMapLayer(
            name="loot",
            columns=self.columns,
            rows=self.rows,
            pixel_width=self.columns * self.tile_size,
            pixel_height=self.rows * self.tile_size,
            sprite_filename=resolve_resource_path("images/star.png"),
            sprite_class=Loot
        )

    def possible_loot_locations(self):
        """Returns a list of points where loot could be placed. 
        """
        X = range(1, self.columns, 2)
        Y = range(1, self.rows, 2)
        return list(product(X, Y))

if __name__ == '__main__':
    game = MazeGame()
    game.run()



from arcade.sprite import Sprite
from arcade import calculate_points
from quest.helpers import scale
import arcade
import os

class QuestSprite(Sprite):
    """
    The base class for sprites in Quest.

    A :py:class:`QuestSprite` is a subclass of :py:class:`arcade.Sprite` with a
    few additional methods to help integrate into the Quest framework.

    Attributes:
        description: A string description of the sprite.
        strategy: If set, should be a :py:class:`quest.strategy.Strategy`, or another class
            instance with a :py:meth:`choose_course` method.
        speed: The sprite's speed.

    Arguments:
        filename: The only required argument is the name of the sprite's image file.
        kwargs: There are many optional keyword arguments inherited from :py:class:`arcade.Sprite`.
    """
    description = "quest sprite"
    strategy = None
    speed = 1

    def set_course(self, vector):
        """Update the `change_x` and `change_y` properties using a vector.

        Normally, sprites' intended movement is set using the `change_x` and `change_y`
        properties. This method updates `change_x` and `change_y` based on an (x, y) vector
        whose magnitude is scaled to the sprite's speed. This is helpful when you have an
        (x, y) vector, for example a :py:class:`quest.helpers.Direction` or a vector to another
        sprite, and want to have this sprite head that direction at its speed. For example, if you wanted
        a sprite to always try to move down at its speed::

                def on_update(self):
                    self.set_course(Direction.DOWN.to_vector())

        """
        vx, vy = scale(vector, self.speed)
        self.change_x = vx
        self.change_y = vy

    def stop(self):
        """Set the sprite's `change_x` and `change_y` to zero.
        """
        self.change_x = 0
        self.change_y = 0

    def on_collision(self, other_sprite, game):
        """Called when the sprite collides with another sprite.

        Override this method to change the sprite's collision behavior.
        For example, an NPC sprite which represents an item that can be collected
        by the player would probably call a method on `game` to update the inventory
        and then call :py:meth:`self.kill()`.
        """
        pass

    def on_update(self, game):
        """Called on every tick, performs any needed updates.

        By default, if the sprite has a strategy, it uses the strategy
        to set its course.
        """
        if self.strategy:
            self.set_course(self.strategy.choose_course(self, game))

    def __str__(self):
        return "<{}>".format(self.description)


class Wall(QuestSprite):
    """Wall.
    """
    description = "wall"

class Player(QuestSprite):
    """Player
    """
    description = "player"

class NPC(QuestSprite):
    """Non-playable character
    """
    description = "npc"

class Background(QuestSprite):
    """A sprite that does nothing.
    """
    description = "background"



# sprite_directionality.py
# by Team Sonic
#
# This file implements a sprite mixin which causes the sprite's texture
# direction to change based on the last direction the sprite was moving.

from quest.sprite import Player
from quest.examples.grandmas_soup import GrandmasSoupGame, Grandma, Carrots, Mushroom, Potatoes, Tomatos
from quest.contrib.sprite_directionality import SpriteDirectionMixin
from quest.helpers import scale, resolve_resource_path
from quest.strategy import RandomWalk
import arcade

class DirectionalPlayer(SpriteDirectionMixin, Player):
    """ Uses the SpriteDirectionMixin to make the player sprite face different directions
    based on which direction the player is moving
    """

class DirectionalGrandma(SpriteDirectionMixin, Grandma):
    """ Uses the SpriteDirectionMixin to make the player sprite face different directions
    based on which direction the player is moving
    """

class DirectionalGame(GrandmasSoupGame):
    """Extends the GrandmasSoupGame to have a player and grandmother than face different
    directions depending on the dirtection they are moving.
    """

    player_sprite_image = resolve_resource_path("images/people/trixie.png")
    player_scaling = 3
    def setup_player(self):
        """Creates the player sprite.

        Initializes a sprite for the player, assigns its starting position,
        and appends the player sprite to a SpriteList (Arcade likes to work
        with sprites in SpriteLists).
        """
        self.player = DirectionalPlayer(self.player_sprite_image, self.player_sprite_image, self.player_sprite_image, self.player_scaling)
        self.player.center_x = self.player_initial_x
        self.player.center_y = self.player_initial_y
        self.player.speed = self.player_speed
        self.player_list = arcade.SpriteList()
        self.player_list.append(self.player)

    def setup_npcs(self):
        """Creates and places Grandma and the vegetables.
        """
        static_npc_data = [
            [Carrots, "images/items/carrots.png", 1, 220, 640],
            [Mushroom, "images/items/mushroom.png", 1, 1028, 264],
            [Potatoes, "images/items/potatoes.png", 1, 959, 991],
            [Tomatos, "images/items/tomatos.png", 1, 323, 1055],
        ]
        self.npc_list = arcade.SpriteList()
        grandma_image = resolve_resource_path("images/people/grandma.png")
        grandma = DirectionalGrandma(grandma_image, grandma_image, grandma_image, 3)
        grandma.center_x = 400
        grandma.center_y = 400
        self.npc_list.append(grandma)
        for sprite_class, image, scale, x, y in static_npc_data:
            sprite = sprite_class(resolve_resource_path(image), scale)
            sprite.center_x = x
            sprite.center_y = y
            self.npc_list.append(sprite)

        grandma = self.npc_list[0]
        walk = RandomWalk(0.05)
        grandma.strategy = walk

if __name__ == "__main__":
    game = DirectionalGame()
    game.run()



from quest.examples.island import IslandAdventure
from quest.contrib.mouse import MouseMotionMixin

class IslandWithMouse(MouseMotionMixin, IslandAdventure):
    "The island adventure with mouse control."

if __name__ == '__main__':
    game = IslandWithMouse()
    game.run()



import arcade
from quest.engines import ContinuousPhysicsEngine
from quest.errors import NoMapError, NoLayerError
from quest.sprite import Player
from time import time


class QuestGame(arcade.Window):
    """Implements a top-down video game with a character on a map.

    :py:class:`QuestGame` is the central class in the :doc:`../index`,
    which is built on top of :doc:`arcade:index`. :py:class:`QuestGame`
    is a subclass of :py:class:`arcade:arcade.Window`.
    To create your own game, create a subclass of :py:class:`QuestGame`
    and then override whatever you need to change from the default behavior.

    When :py:class:`QuestGame` is initialized, it sets up the player, the maps,
    the walls, the NPCs, the physics engine, and centers the viewport on the player.
    Rather than overriding :py:meth:`__init__`, consider overriding just the setup
    functions you need to change.

    Attributes:
        screen_width: Width in pixels of the game window.
        screen_height: Height in pixels of the game window.
        left_viewport_margin: Minimum distance (in pixels) between the
            player sprite and the left edge of the viewport.
        right_viewport_margin: Minimum distance (in pixels) between the
            player sprite and the right edge of the viewport.
        bottom_viewport_margin: Minimum distance (in pixels) between the
            player sprite and the bottom edge of the viewport.
        top_viewport_margin: Minimum distance (in pixels) between the
            player sprite and the top edge of the viewport.
        screen_title: Title of the game window (displayed top center)
        player_scaling: Factor by which to scale the player sprite.
        player_sprite_image: Filepath for the player sprite.
        player_speed: In pixels per update. (By default, the game runs
            at 60 hertz, so update is called every 1/60 second.)
        player_initial_x: Initial x-coordinate for player center.
        player_initial_y: Initial y-coordinate for player center.
        view_bottom: y-coordinate of the bottom edge of the current viewport
        view_left: x-coordinate of the left edge of the current viewport
    """
    screen_width = 600
    screen_height = 600
    left_viewport_margin = 100
    right_viewport_margin = 100
    bottom_viewport_margin = 100
    top_viewport_margin = 100
    screen_title = "Quest"
    player_scaling = 1
    player_sprite_image = None
    player_speed = 10
    player_initial_x = 0
    player_initial_y = 0
    view_bottom = 0
    view_left = 0

    def __init__(self):
        """Initializes the game window and sets up other classes.
        """
        super().__init__(self.screen_width, self.screen_height, self.screen_title)
        self.running = False
        self.setup_maps()
        if len(self.maps) > 0:
            self.set_current_map(0)
        self.setup_player()
        self.setup_walls()
        self.setup_npcs()
        self.setup_physics_engine()
        self.center_view_on_player()
        self.current_modal = None

    def run(self):
        """Starts the game.
        """
        self.start_time = time()
        self.running = True
        arcade.run()

    def setup_maps(self):
        """Sets up the game maps.

        self.maps should be assigned to a list of :py:class:`Map` objects, which get
        initialized here. Each map represents a 'level' or 'scene' of the game.
        Once the list of maps is created, use :py:meth:`set_current_map` to
        set one of the maps as the initial current map.
        This method will need to be overridden by any game using a map.
        For more details, see :ref:`creating_maps`
        """
        self.maps = []

    def setup_player(self):
        """Creates the player sprite.

        Initializes a sprite for the player, assigns its starting position,
        and appends the player sprite to a SpriteList (Arcade likes to work
        with sprites in SpriteLists).
        """
        self.player = Player(self.player_sprite_image, self.player_scaling)
        self.player.center_x = self.player_initial_x
        self.player.center_y = self.player_initial_y
        self.player.speed = self.player_speed
        self.player_list = arcade.SpriteList()
        self.player_list.append(self.player)

    def setup_walls(self):
        """Does any neccessary setup for NPCs.
        """
        self.wall_list = arcade.SpriteList()

    def setup_npcs(self):
        """Does any neccessary setup for NPCs.
        """
        self.npc_list = arcade.SpriteList()

    def add_map(self, game_map):
        """Adds a map to the list of maps.

        Arguments:
            game_map: A quest.map.Map object.
        """
        self.maps.append(game_map)
        if len(self.maps) == 1:
            self.set_current_map(0)

    def get_current_map(self):
        """Gets the current game map.

        The current map is tracked using ``current_map_index``.

        Returns:
            The current Map.
        """
        if self.current_map_index is None:
            raise NoMapError("The game has no maps")
        return self.maps[self.current_map_index]

    def set_current_map(self, index):
        """Sets the current game map.

        Checks to make sure ``index`` is valid, and then stores it as
        ``current_map_index``.
        """
        if index < 0 or index >= len(self.maps):
            raise ValueError("Cannot set current map to {}; there are {} maps.".format(
                    index, len(self.maps)))
        self.current_map_index = index

    def setup_physics_engine(self):
        """Sets up the physics engine.

        Initializes the physics engine that will be used in the game.
        A physics engine resolves interactions between sprites according to a
        set of rules. The :doc:`arcade:index` :py:class:`arcade:PhysicsEngineSimple`
        just keeps the player sprite from bumping into walls. More complicated
        physics engines could implement collisions, gravity, or even realistic
        3-dimensional interactions.

        By default, :py:class:`QuestGame` uses :py:class:`arcade:PhysicsEngineSimple`
        to prevent the player sprite from colliding or passing through any
        sprites on a map layer with the ``wall`` role. If there are no
        :py:class:`MapLayer` with the ``wall`` role, uses the
        :py:class:`NullPhysicsEngine` instead. Don't override this method
        unless you understand Arcade's physics engines pretty well.
        """
        self.physics_engine = ContinuousPhysicsEngine(self)

    def on_update(self, delta_time):
        """Updates the game's state.

        At every tick, the game needs to be updated. The physics engine
        updates sprite positions, and then sprite callbacks are executed.
        Finally, the viewport is scrolled. Note that `on_update` changes the
        state of the game, but does not draw anything to the screen.

        Args:
            delta_time: How much time has passed since the last update.
        """
        if self.running:
            self.player.on_update(self)
            for npc in self.npc_list:
                npc.on_update(self)
            self.physics_engine.update()
            self.scroll_viewport()

    def on_draw(self):
        """Draws the screen.

        At every tick, just after `on_update`, the whole screen needs to be
        redrawn. This involves drawing the background color, each map layer
        with the `display` role, the NPC's, the player, and any message that
        needs to be displayed.
        """
        arcade.start_render()
        arcade.set_background_color(self.get_current_map().background_color)
        for layer in self.get_current_map().layers:
            layer.draw()
        self.npc_list.draw()
        self.player_list.draw()
        message = self.message()
        if message:
            arcade.draw_text(message, 10 + self.view_left, 10 + self.view_bottom,
                    arcade.csscolor.WHITE, 18)
        if self.current_modal:
            self.current_modal.on_draw()

    def open_modal(self, modal):
        """Shows a modal window and pauses the game until the modal resolves.
        """
        self.running = False
        self.current_modal = modal
        self.current_modal.update_position(
            self.view_left + self.screen_width / 2,
            self.view_bottom + self.screen_height / 2,
        )

    def close_modal(self):
        """Resolves a modal window and resumes the game.
        """
        self.current_modal = None
        self.running = True

    def on_key_press(self, key, modifiers):
        """Handles key presses.

        Arguments:
            key: The key that was pressed.
            modifiers: A list of currently-active modifier keys (e.g. shift).

        While a key is pressed, the sprite's x- and y- change values are set
        to the player's movement speed. Think of this as an intention to move;
        it's up to the physics engine to decide whether this actually
        results in movement. For example, the physics engine will prevent
        players from moving into walls. This method is automatically called at
        the appropriate time.
        """
        if self.current_modal:
            self.current_modal.on_key_press(key, modifiers)
        else:
            if key == arcade.key.UP or key == arcade.key.W:
                self.player.change_y = self.player.speed
            elif key == arcade.key.DOWN or key == arcade.key.S:
                self.player.change_y = -self.player.speed
            if key == arcade.key.LEFT or key == arcade.key.A:
                self.player.change_x = -self.player.speed
            elif key == arcade.key.RIGHT or key == arcade.key.D:
                self.player.change_x = self.player.speed

    def on_key_release(self, key, modifiers):
        """Handles key releases.

        Arguments:
            key: The key that was released.
            modifiers: A list of currently-active modifier keys (e.g. shift).

        Whenever a key is released, the player's change_x or change_y
        is set to 0, indicating that the player no longer intends to keep
        moving. This method is automatically called at the appropriate time.
        """
        if self.current_modal:
            self.current_modal.on_key_release(key, modifiers)
        else:
            if key == arcade.key.UP or key == arcade.key.W:
                self.player.change_y = 0
            elif key == arcade.key.DOWN or key == arcade.key.S:
                self.player.change_y = 0
            elif key == arcade.key.LEFT or key == arcade.key.A:
                self.player.change_x = 0
            elif key == arcade.key.RIGHT or key == arcade.key.D:
                self.player.change_x = 0

    def center_view_on_player(self):
        """Centers the viewport on the player.
        """
        self.view_left = self.player.center_x - self.screen_width / 2
        self.view_bottom = self.player.center_y - self.screen_height / 2
        self.update_viewport()
        self.scroll_viewport()

    def scroll_viewport(self):
        """Updates the viewport to keep the player within margins.

        If the player sprite is too close to any edge of the viewport
        (or has somehow gone beyond the viewport), scrolls the viewport
        to the player. The {left, right, bottom, top}_viewport_margin
        properties specify how close the player is allowed to be to the
        edge before scrolling.
        """

        changed = False

        left_boundary = self.view_left + self.left_viewport_margin
        if self.player.left < left_boundary:
            self.view_left -= left_boundary - self.player.left
            changed = True

        right_boundary = self.view_left + self.screen_width - self.right_viewport_margin
        if self.player.right > right_boundary:
            self.view_left += self.player.right - right_boundary
            changed = True

        top_boundary = self.view_bottom + self.screen_height - self.top_viewport_margin
        if self.player.top > top_boundary:
            self.view_bottom += self.player.top - top_boundary
            changed = True

        bottom_boundary = self.view_bottom + self.bottom_viewport_margin
        if self.player.bottom < bottom_boundary:
            self.view_bottom -= bottom_boundary - self.player.bottom
            changed = True

        if changed:
            self.view_bottom = int(self.view_bottom)
            self.view_left = int(self.view_left)
            self.update_viewport()

    def update_viewport(self):
        """Updates the viewport.

        Uses the `view_left`, `view_bottom`, and the screen size
        properties to update the viewport. Needs to be called after
        changing any of these properties.
        """
        arcade.set_viewport(
            self.view_left,
            self.screen_width + self.view_left,
            self.view_bottom,
            self.screen_height + self.view_bottom
        )

    def message(self):
        """Generates a message (or no message) to be shown on screen.

        Returns:
            A string to be shown on screen, or None if no message is needed.
        """
        return None



from PIL import Image
import xml.etree.ElementTree as ET
import quest
from itertools import product, chain
from pathlib import Path
from enum import Flag, auto
from math import sqrt
import re

def tint(color, ratio=0.25):
    """Creates a tint of a color by scaling it toward pure white.

    Arguments:
        color (int, int, int): The base color.
        ratio (float): A value between 0 and 1. 0 would have no effect; 
            1 would be pure white.
    """
    return tuple(round(255 - (255 - c) * (1 - ratio)) for c in color)

def shade(color, ratio=0.25):
    """Creates a shade of a color by scaling it toward pure black.

    Arguments:
        color (int, int, int): The base color.
        ratio (float): A value between 0 and 1. 0 would have no effect; 
            1 would be pure black.
    """
    return tuple(round(c * (1 - ratio)) for c in color)

class Direction(Flag):
    """Direction lets you talk about directions like `Direction.DOWN`, `Direction.UPLEFT`
    or using compass directions such as `Direction.NE`.
    """

    NONE = 0
    RIGHT = auto()
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    UPRIGHT = UP | RIGHT
    UPLEFT = UP | LEFT
    DOWNLEFT = DOWN | LEFT
    DOWNRIGHT = DOWN | RIGHT

    E = RIGHT
    N = UP
    S = DOWN
    W = LEFT
    NE = N | E
    NW = N | W
    SW = S | W
    SE = S | E

    PRIMARY_AXIS = LEFT | RIGHT

    @classmethod
    def from_vector(cls, vector, diagonal=True):
        """Converts an (x, y) tuple into a direction. 

            >>> Direction.from_vector((-1, 0))
            Direction.LEFT
            >>> Direction.from_vector((0.4, 0.6))
            Direction.UPRIGHT

        Arguments: 
            vector (float, float): An (x, y) tuple.
            diagonal (bool): Whether to include diagonal directions. Defaults to True.

        Returns:
            A Direction.
        """
        vx, vy = vector
        result = cls.NONE
        if vx < 0: 
            result |= cls.LEFT
        if vx > 0: 
            result |= cls.RIGHT
        if vy < 0:
            result |= cls.DOWN
        if vy > 0:
            result |= cls.UP
        if result.is_diagonal() and not diagonal:
            result &= cls.PRIMARY_AXIS
        return result

    def is_diagonal(self):
        """Returns whether the Direction is diaognal.

            >>> Direction.NW.is_diagonal()
            True
        """
        return self in [self.NE, self.NW, self.SW, self.SE]

    def turn_clockwise(self):
        """Returns the direction 90 degrees clockwise. Only supports cardinal directions.
        """
        return {
            Direction.E: Direction.S,
            Direction.S: Direction.W,
            Direction.W: Direction.N,
            Direction.N: Direction.E,
        }[self]

    def turn_anticlockwise(self):
        """Returns the direction 90 degrees anticlockwise. Only supports cardinal directions.
        """
        return {
            Direction.E: Direction.N,
            Direction.N: Direction.W,
            Direction.W: Direction.S,
            Direction.S: Direction.E,
        }[self]

    def to_vector(self, normalized=False):
        """Converts the Direction into an (x, y) tuple.

        Arugments: 
            normalized (bool): Whether to normalize the vector so that its 
                magintude is 1. Defaults to False, in which case x and y are
                each either 0 or 1.
        """
        vx, vy = 0, 0
        if self & self.RIGHT:
            vx += 1
        if self & self.UP:
            vy += 1
        if self & self.LEFT:
            vx -= 1
        if self & self.DOWN:
            vy -= 1
        if normalized:
            vx, vy = normalize((vx, vy))
        return vx, vy

class SpriteListList:
    """Allows multiple SpriteLists to be treated as if they were a single SpriteList.
    """
    def __init__(self, sprite_lists):
        self.sprite_lists = sprite_lists

    def chain_sprite_lists(self):
        return chain.from_iterable(self.sprite_lists)

    def __iter__(self):
        return iter(self.chain_sprite_lists())

    def update(self):
        for sprite_list in self.chain_sprite_lists():
            sprite_list.update()

def tileset_to_collection(image_path, tile_size, output_dir, name="tileset", create_tsx=True):
    """Splits a tileset image into separate files.

    Quest relies on Arcade, which only works with collections of images. 
    A lot of game art is provided as a single tile image. This function can help split
    it out.

    Args:
        image_path: Path to an image containing a grid of tiles.
        tile_size: Size in pixels of each tile.
        output_dir: Directory for output files (there may be a lot of them).
        name: Name of the tileset.
        create_tsx: If True, also creates a tsx file which can be opened as a
            :doc:`Tileset <tiled:manual/editing-tilesets>` using 
            :doc:`Tiled <tiled:manual/introduction>`.

    """
    img = Image.open(image_path)
    width, height = img.size
    cols, rows = width // tile_size, height // tile_size
    for i, j in product(range(cols), range(rows)):
        left = i * tile_size
        upper = j * tile_size
        right = left + tile_size
        lower = upper + tile_size
        tile = img.crop((left, upper, right, lower))
        output_path = Path(output_dir) / "img_{}_{}.png".format(j, i)
        tile.save(output_path)

    if create_tsx:
        tileset = empty_tileset(tile_size, cols, rows, name)
        for ix, (j, i) in enumerate(product(range(rows), range(cols))):
            add_tile(tileset, ix, "img_{}_{}.png".format(j, i))
        with open(Path(output_dir) / "{}.tsx".format(name), 'wb') as f:
            ET.ElementTree(tileset).write(f, encoding="UTF-8", xml_declaration=True)

def empty_tileset(tile_size, cols, rows, name):
    tileset = ET.Element("tileset")
    tileset.set('version', "1.2")
    tileset.set('tiledversion',"1.3.1")
    tileset.set('name', name)
    tileset.set('tilewidth', str(cols))
    tileset.set('tileheight', str(rows))
    tileset.set('tilecount', str(cols * rows))
    tileset.set('columns', "0")
    grid = ET.SubElement(tileset, "grid")
    grid.set('orientation', "orthagonal")
    grid.set('width', "1")
    grid.set('height', "1")
    return tileset

def add_tile(tileset, tile_id, img_path):
    tile_element = ET.SubElement(tileset, 'tile')
    tile_element.set('id', str(tile_id) )
    img_element = ET.SubElement(tile_element, 'image')
    img_element.set('width', tileset.get('tilewidth'))
    img_element.set('height', tileset.get('tileheight'))
    img_element.set('source', img_path)

def normalize(vector):
    return scale(vector, 1)

def scale(vector, magnitude):
    vx, vy = vector
    old_magnitude = sqrt(vx * vx + vy * vy) if vx * vx + vy * vy else 0
    factor = magnitude / old_magnitude
    return vx * factor, vy * factor
    
class SimpleInkParser:
    """Parses a simple subset of Ink syntax into a JSON-like data structure. 

    The ink must meet the following constraints:

    - Must be valid Ink.
    - All content must be in a knot. Knots must be delimited 
      with three equal signs on either side of the knot name.
    - The only syntax allowed is knot declarations, sticky choices
      (+) and diverts (->). Diverts are only allowed following a sticky
      choice.
    """
    def parse(self, ink):
        """Reads a story written in a subset of Ink syntax (described above) and returns a
        data structure of content and choices, also described above.
        """
        dialogue = {}
        knots = self.split_knots(ink)
        for line_num, knot_name, knot_ink in knots:
            content, options = self.parse_knot_ink(line_num, knot_ink)
            if len(options) == 0:
                raise ValueError("line {}: Knot {} has no options".format(line_num, knot_name))
            if knot_name in dialogue.keys():
                raise ValueError("line {}: Knot {} already defined".format(line_num, knot_name))
            dialogue[knot_name] = {"content": content, "options": options}
        return dialogue

    def parse_knot_ink(self, line_num, ink):
        """Reads lines of code in a knot and returns a list of content and a dict of options.
        """
        content, options = self.split_content_from_options(ink)
        content = [c.strip() for c in content]
        content = self.split_and_join(content, lambda c: not c)
        parsed_content = [c.strip() for c in content if c.strip()]
        options = [o.strip() for o in options]
        options = self.split_and_join(options, lambda o: re.match("\s*\+", o))
        parsed_options = {}
        for option in options[1:]: 
            match = re.match("\s*\+(?P<text>.*)\->\s*(?P<knot>[a-zA-Z_]+)", option)
            if not match: 
                raise ValueError("line {}: Error reading option in knot.".format(line_num))
            parsed_options[match.group('text').strip()] = match.group('knot')
        return parsed_content, parsed_options

    def split_and_join(self, strings, condition):
        """Splits a list of strings on a condition and joins the results. For example, 

            >>> vowel = lambda l: l in 'aeiou'
            >>> split_and_join('abcdefghijklmnop', vowel)
            ['a', 'bcde', 'fghi', 'jklmno', 'p']
        """
        splits = [i for i, s in enumerate(strings) if condition(s)]
        return [' '.join(strings[i:j]) for i, j in zip([0] + splits, splits + [len(strings)])]

    def split_content_from_options(self, ink):
        content, options = [], []
        for line in ink:
            if any(options) or re.match("\s*\+", line):
                options.append(line)
            else:
                content.append(line)
        return content, options

    def split_knots(self, ink):
        knots = []
        current_name = None
        current_ink = []
        for i, line in enumerate(ink):
            name = self.parse_knot_declaration(line)
            # The first knot has not started yet 
            if name is None and current_name is None:
                continue
            # Starting the first knot
            elif name is not None and current_name is None:
                current_name = name
            # Starting a new knot
            elif name is not None:
                if len(current_ink) == 0:
                    raise ValueError("Expected knot content at line {}".format(i))
                starting_line_num = i - len(current_ink)
                knots.append((starting_line_num, current_name, current_ink))
                current_name = name
                current_ink = []
            # Continuing inside a knot
            else:
                current_ink.append(line)
        # At end of file, the final knot's contents
        if len(current_ink) == 0:
            raise ValueError("Expected knot content at line {}".format(i))
        starting_line_num = i - len(current_ink) - 1
        knots.append((starting_line_num, current_name, current_ink))
        return knots
                
    def parse_knot_declaration(self, line):
        "Returns knot name if found"
        match = re.match("===\s+([a-zA-Z_]+)\s+===", line)
        return match.group(1) if match else None

def resolve_resource_path(filename):
    "Resolves a resource path for example game resouces (so they can be played from anywhere)."
    return str(Path(quest.__file__).parent / "examples" / filename)



import setuptools
from pathlib import Path

setuptools.setup(
    name="questgame",
    version="0.1.6",
    author="Chris Proctor",
    author_email="chris@chrisproctor.net",
    description="A simple game framework",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/cproctor/quest",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "arcade==2.3.8",
        "easing-functions>=1.0.3",
        "numpy>=1.17.4",
        "Pillow>=6.2.1",
        "tqdm>=4.42.1",
        "xvfbwrapper>=0.2.9",
    ]
)



from quest.examples.island import IslandAdventure
from quest.contrib.player_enabled import PlayerEnabledMixin
from datetime import datetime

class SleepyIslandAdventure(PlayerEnabledMixin, IslandAdventure):
    """The Island Adventure, but in this version the player occasionally needs to sleep.
    """

    def on_update(self, delta_time):
        if self.is_sleeping():
            self.player_enabled = False
            self.message_str = "Sleeping..."
        else:
            self.player_enabled = True
            self.message_str = ""
        super().on_update(delta_time)


    def is_sleeping(self):
        """Determines whether the player is sleeping. Alternates every three seconds.
        """
        return datetime.now().second % 6 < 3

    def message(self):
        return self.message_str
    
if __name__ == '__main__':
    game = SleepyIslandAdventure()
    game.run()



# Modal
# -------------------
# A modal window is a "pop-up" that creates a mode--it pauses the game until the 
# window

import arcade
from quest.text_label import TextLabelStack

class Modal:
    """ A modal window is a pop-up that pauses the game until it is resolved.
    It is called a Modal because it puts the game into a new mode. 

    Arguments: 
        game: The game which the Modal is part of.
    """
    width = 400
    height = 400
    background_color = arcade.color.LIGHT_GRAY

    def __init__(self, game):
        self.game = game
        self.x_center = 0
        self.y_center = 0
        self.set_text_labels()
        self.set_option_labels()

    def update_position(self, x, y):
        self.x_center = x
        self.y_center = y
        self.set_text_labels()
        self.set_option_labels()

    def set_text_labels(self):
        self.text_labels = TextLabelStack(
            ["Sorry to bother you with this modal.", "But it's important."], 
            self.x_center,
            self.y_center + self.height / 2
        )

    def set_option_labels(self):
        self.option_labels = TextLabelStack(
            ["OK", "Fine"],
            self.x_center,
            self.y_center + 100
        )
        self.option_labels.set_highlight(0)

    def choose_option(self, value):
        """When a button is clicked, it calls choose_option with its value.
        """
        self.game.close_modal()

    def on_key_release(self, key, modifiers):
        pass

    def on_key_press(self, key, modifiers):
        if key == arcade.key.UP or key == arcade.key.W:
            i = (self.option_labels.get_highlight() + 1) % len(self.option_labels)
            self.option_labels.set_highlight(i)
        elif key == arcade.key.DOWN or key == arcade.key.S:
            i = (self.option_labels.get_highlight() - 1) % len(self.option_labels)
            self.option_labels.set_highlight(i)
        elif key == arcade.key.ENTER:
            self.choose_option(self.option_labels.get_highlight())

    def on_draw(self):
        arcade.draw_rectangle_filled(
            self.x_center, 
            self.y_center, 
            self.width, 
            self.height, 
            self.background_color
        )
        self.text_labels.draw()
        self.option_labels.draw()

class DialogueModal(Modal):
    """A modal window powered by a Dialogue object.
    """
    def __init__(self, game, dialogue):
        self.dialogue = dialogue
        super().__init__(game)

    def set_text_labels(self):
        self.text_labels = TextLabelStack(
            self.dialogue.get_content(),
            self.x_center,
            self.y_center + self.height / 2
        )

    def set_option_labels(self):
        self.option_labels = TextLabelStack(
            self.dialogue.get_options(),
            self.x_center,
            self.y_center 
        )
        self.option_labels.set_highlight(0)

    def choose_option(self, value):
        self.dialogue.choose(value)
        if self.dialogue.running:
            self.set_text_labels()
            self.set_option_labels()
        else:
            self.game.close_modal()



from quest.examples.maze import MazeGame
from quest.contrib.maze_walk_strategy import MazeWalkStrategy
from quest.sprite import NPC
from quest.helpers import resolve_resource_path
from datetime import datetime

class MazeWalkerGame(MazeGame):
    """Extends the MazeGame to add an evil NPC who steals loot. 
    """

    def setup_npcs(self):
        super().setup_npcs()
        enemy = NPC(resolve_resource_path("images/people/pops.png"), 1.5)
        grid = self.get_current_map().get_layer_by_name("walls")
        enemy.strategy = MazeWalkStrategy(grid, self.wall_list)
        enemy.center_x = self.player.center_x
        enemy.center_y = self.player.center_y
        enemy.speed = 4
        self.npc_list.append(enemy)

    def on_loot_collected(self, collector):
        """A method to be called whenever loot is collected.

        See the :py:class:`Loot` NPC sprite below. It calls this method whenever there is
        a collision with the player.
        """
        if collector == self.player:
            self.score += 1
        else:
            self.max_score -= 1
        if self.score == self.max_score:
            self.game_over = True
            self.final_elapsed_time = (datetime.now() - self.level_start).seconds
    
    def message(self):
        """Returns a string like "Time 12 (12/25)"
        """
        if self.game_over:
            return "Final score: {} in {} seconds".format(self.max_score, self.final_elapsed_time)
        else:
            elapsed_time = datetime.now() - self.level_start
            return "Time {} ({}/{})".format(elapsed_time.seconds, self.score, self.max_score)

if __name__ == '__main__':
    game = MazeWalkerGame()
    game.run()
